import logging
import unittest

import numpy as np
import torch
from sklearn.model_selection import RandomizedSearchCV

from birder.eval.methods.knn import evaluate_knn
from birder.eval.methods.mlp import evaluate_mlp
from birder.eval.methods.mlp import train_mlp
from birder.eval.methods.simpleshot import evaluate_simpleshot
from birder.eval.methods.svm import evaluate_svm
from birder.eval.methods.svm import train_svm

logging.disable(logging.CRITICAL)


def _make_separable_data() -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    rng = np.random.default_rng(0)
    class0 = rng.normal(loc=-1.0, scale=0.1, size=(10, 2)).astype(np.float32)
    class1 = rng.normal(loc=1.0, scale=0.1, size=(10, 2)).astype(np.float32)

    train_features = np.vstack([class0[:8], class1[:8]])
    train_labels = np.array([0] * 8 + [1] * 8, dtype=np.int_)

    test_features = np.vstack([class0[8:], class1[8:]])
    test_labels = np.array([0] * 2 + [1] * 2, dtype=np.int_)

    return (train_features, train_labels, test_features, test_labels)


class TestMethods(unittest.TestCase):
    def test_train_svm_returns_search(self) -> None:
        train_features, train_labels, _, _ = _make_separable_data()
        svc = train_svm(train_features, train_labels, n_iter=2, n_jobs=1, seed=0)

        self.assertIsInstance(svc, RandomizedSearchCV)
        self.assertIsNotNone(svc.best_estimator_)
        self.assertIn("svc__C", svc.best_params_)

        pipeline = svc.best_estimator_
        self.assertEqual([name for name, _ in pipeline.steps], ["standardscaler", "svc"])

    def test_evaluate_svm_predictions(self) -> None:
        train_features, train_labels, test_features, test_labels = _make_separable_data()
        y_pred, y_true = evaluate_svm(
            train_features,
            train_labels,
            test_features,
            test_labels,
            n_iter=2,
            n_jobs=1,
            seed=0,
        )

        self.assertTrue(np.array_equal(y_true, test_labels))
        self.assertEqual(y_pred.shape, test_labels.shape)
        self.assertTrue(np.isin(y_pred, [0, 1]).all())

    def test_evaluate_simpleshot_predictions(self) -> None:
        train_features, train_labels, test_features, test_labels = _make_separable_data()
        y_pred, y_true = evaluate_simpleshot(train_features, train_labels, test_features, test_labels)

        self.assertTrue(np.array_equal(y_true, test_labels))
        self.assertEqual(y_pred.shape, test_labels.shape)
        self.assertTrue(np.isin(y_pred, [0, 1]).all())
        self.assertEqual(float(np.mean(y_pred == y_true)), 1.0)

    def test_evaluate_knn_predictions(self) -> None:
        train_features, train_labels, test_features, test_labels = _make_separable_data()
        y_pred, y_true = evaluate_knn(
            train_features, train_labels, test_features, test_labels, k=3, device=torch.device("cpu")
        )

        self.assertTrue(np.array_equal(y_true, test_labels))
        self.assertEqual(y_pred.shape, test_labels.shape)
        self.assertTrue(np.isin(y_pred, [0, 1]).all())
        self.assertEqual(float(np.mean(y_pred == y_true)), 1.0)

    def test_evaluate_knn_with_k_larger_than_train_set(self) -> None:
        train_features, train_labels, test_features, test_labels = _make_separable_data()
        y_pred, y_true = evaluate_knn(train_features, train_labels, test_features, test_labels, k=100)

        self.assertTrue(np.array_equal(y_true, test_labels))
        self.assertEqual(y_pred.shape, test_labels.shape)
        self.assertTrue(np.isin(y_pred, [0, 1]).all())

    def test_mlp_probe_predictions(self) -> None:
        rng = np.random.default_rng(0)

        # Create separable multi-label data
        n_train = 50
        n_test = 10
        n_features = 8
        n_classes = 2

        train_features = rng.normal(size=(n_train, n_features)).astype(np.float32)
        train_labels = (train_features[:, :n_classes] > 0).astype(np.float32)

        test_features = rng.normal(size=(n_test, n_features)).astype(np.float32)
        test_labels = (test_features[:, :n_classes] > 0).astype(np.float32)

        device = torch.device("cpu")
        model = train_mlp(
            train_features,
            train_labels,
            num_classes=n_classes,
            device=device,
            epochs=10,
            batch_size=8,
            hidden_dim=64,
            seed=0,
        )

        y_pred, y_true, macro_f1 = evaluate_mlp(model, test_features, test_labels, device=device)

        self.assertEqual(y_pred.shape, (n_test, n_classes))
        self.assertEqual(y_true.shape, (n_test, n_classes))
        self.assertTrue(np.isin(y_pred, [0, 1]).all())
        self.assertGreaterEqual(macro_f1, 0.0)
        self.assertLessEqual(macro_f1, 1.0)



import logging
import unittest

import torch
import webdataset as wds

from birder.data.dataloader.webdataset import make_wds_loader

logging.disable(logging.CRITICAL)


class TestWdsLoader(unittest.TestCase):
    def test_wds_loader_infinite_mode(self) -> None:
        n_samples = 50
        batch_size = 10

        # Create synthetic data: yields (tensor, label)
        mock_ds = wds.MockDataset(sample=(torch.randn(5), torch.tensor(1)), length=n_samples)
        dataset = wds.DataPipeline(mock_ds).with_length(n_samples, silent=True)

        # Infinite dataloader
        loader = make_wds_loader(
            dataset=dataset,
            batch_size=batch_size,
            num_workers=0,
            prefetch_factor=None,
            collate_fn=None,
            world_size=1,
            pin_memory=False,
            drop_last=False,
            shuffle=False,
            exact=False,
            infinite=True,
        )

        self.assertEqual(len(loader), 5)

        iterator = iter(loader)
        batches_fetched = 0
        try:
            for _ in range(15):
                next(iterator)
                batches_fetched += 1
        except StopIteration:
            self.fail("Infinite loader stopped iterating prematurely")

        self.assertEqual(batches_fetched, 15)

    def test_wds_loader_exact_mode(self) -> None:
        n_samples = 23
        batch_size = 10

        # Create synthetic data: yields (tensor, label)
        mock_ds = wds.MockDataset(sample=(torch.randn(5), torch.tensor(1)), length=n_samples)
        dataset = wds.DataPipeline(mock_ds).with_length(n_samples, silent=True)

        # Exact dataloader, partial batch
        loader = make_wds_loader(
            dataset=dataset,
            batch_size=batch_size,
            num_workers=0,
            prefetch_factor=None,
            collate_fn=None,
            world_size=1,
            pin_memory=False,
            drop_last=False,
            shuffle=False,
            exact=True,
            infinite=False,
        )

        self.assertEqual(len(loader), 3)

        batches = list(loader)
        self.assertEqual(len(batches), 3)

        # Check partial batch size
        self.assertEqual(len(batches[-1][0]), 3)

        # Exact dataloader, drop last
        loader = make_wds_loader(
            dataset=dataset,
            batch_size=batch_size,
            num_workers=0,
            prefetch_factor=None,
            collate_fn=None,
            world_size=1,
            pin_memory=False,
            drop_last=True,
            shuffle=False,
            exact=True,
            infinite=False,
        )

        self.assertEqual(len(loader), 2)

        batches = list(loader)
        self.assertEqual(len(batches), 2)

        # Check last batch
        self.assertEqual(len(batches[-1][0]), batch_size)



from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import torch
from PIL import Image


@dataclass(frozen=True)
class InterpretabilityResult:
    original_image: npt.NDArray[np.float32]
    visualization: npt.NDArray[np.float32] | npt.NDArray[np.uint8]
    raw_output: npt.NDArray[np.float32]
    logits: Optional[torch.Tensor] = None
    predicted_class: Optional[int] = None

    def show(self, figsize: tuple[int, int] = (12, 8)) -> None:
        _, (ax1, ax2) = plt.subplots(1, 2, figsize=figsize)
        ax1.imshow(self.visualization)
        ax2.imshow(self.original_image)
        plt.show()


def load_image(image: str | Path | Image.Image) -> Image.Image:
    if isinstance(image, (str, Path)):
        return Image.open(image)

    return image


def preprocess_image(
    image: str | Path | Image.Image, transform: Callable[..., torch.Tensor], device: torch.device
) -> tuple[torch.Tensor, npt.NDArray[np.float32]]:
    pil_image = load_image(image)
    input_tensor = transform(pil_image).unsqueeze(dim=0).to(device)

    # Resize and normalize for visualization
    resized = pil_image.resize((input_tensor.shape[-1], input_tensor.shape[-2]))
    rgb_img = np.array(resized).astype(np.float32) / 255.0

    return (input_tensor, rgb_img)


def show_mask_on_image(
    img: npt.NDArray[np.float32],
    mask: npt.NDArray[np.float32],
    image_weight: float = 0.5,
    colormap: str = "jet",
) -> npt.NDArray[np.uint8]:
    color_map = matplotlib.colormaps[colormap]
    heatmap = color_map(mask)[:, :, :3]

    cam: npt.NDArray[np.float32] = (1 - image_weight) * heatmap + image_weight * img
    cam = cam / np.max(cam)
    cam = cam * 255

    return cam.astype(np.uint8)


def scale_cam_image(
    cam: npt.NDArray[np.float32], target_size: Optional[tuple[int, int]] = None
) -> npt.NDArray[np.float32]:
    result = []
    for img in cam:
        img = img - np.min(img)
        img = img / (1e-7 + np.max(img))
        if target_size is not None:
            img = np.array(Image.fromarray(img).resize(target_size))

        result.append(img)

    return np.array(result, dtype=np.float32)


def deprocess_image(img: npt.NDArray[np.float32]) -> npt.NDArray[np.uint8]:
    img = img - np.mean(img)
    img = img / (np.std(img) + 1e-5)
    img = img * 0.1
    img = img + 0.5
    img = np.clip(img, 0, 1)

    return np.array(img * 255).astype(np.uint8)


def validate_target_class(target_class: Optional[int], num_classes: int) -> None:
    if target_class is not None:
        if target_class < 0 or target_class >= num_classes:
            raise ValueError(f"target_class must be in range [0, {num_classes}), got {target_class}")


def predict_class(logits: torch.Tensor) -> int:
    return int(torch.argmax(logits, dim=-1).item())



from birder.introspection.attention_rollout import AttentionRollout
from birder.introspection.base import InterpretabilityResult
from birder.introspection.feature_pca import FeaturePCA
from birder.introspection.gradcam import GradCAM
from birder.introspection.guided_backprop import GuidedBackprop
from birder.introspection.transformer_attribution import TransformerAttribution

__all__ = [
    "InterpretabilityResult",
    "AttentionRollout",
    "FeaturePCA",
    "GradCAM",
    "GuidedBackprop",
    "TransformerAttribution",
]



from dataclasses import dataclass
from typing import Optional
from typing import Protocol

import torch

from birder.data.transforms.classification import RGBType


@dataclass(frozen=True)
class AttackResult:
    adv_inputs: torch.Tensor
    adv_logits: torch.Tensor
    perturbation: torch.Tensor
    logits: Optional[torch.Tensor] = None
    success: Optional[torch.Tensor] = None
    num_queries: Optional[int] = None


class Attack(Protocol):
    def __call__(self, input_tensor: torch.Tensor, target: Optional[torch.Tensor]) -> AttackResult: ...


def _to_channel_tensor(
    values: tuple[float, float, float], device: Optional[torch.device], dtype: Optional[torch.dtype]
) -> torch.Tensor:
    return torch.tensor(values, device=device, dtype=dtype).view(1, -1, 1, 1)


def normalized_bounds(
    rgb_stats: RGBType, device: Optional[torch.device] = None, dtype: Optional[torch.dtype] = None
) -> tuple[torch.Tensor, torch.Tensor]:
    mean = _to_channel_tensor(rgb_stats["mean"], device=device, dtype=dtype)
    std = _to_channel_tensor(rgb_stats["std"], device=device, dtype=dtype)
    min_val = (0.0 - mean) / std
    max_val = (1.0 - mean) / std

    return (min_val, max_val)


def pixel_eps_to_normalized(
    eps: float | torch.Tensor,
    rgb_stats: RGBType,
    device: Optional[torch.device] = None,
    dtype: Optional[torch.dtype] = None,
) -> torch.Tensor:
    eps_tensor = torch.as_tensor(eps, device=device, dtype=dtype)
    std = _to_channel_tensor(rgb_stats["std"], device=eps_tensor.device, dtype=eps_tensor.dtype)

    if eps_tensor.numel() == 1:
        eps_tensor = eps_tensor.reshape(1, 1, 1, 1)
    else:
        eps_tensor = eps_tensor.reshape(1, -1, 1, 1)

    return eps_tensor / std


def clamp_normalized(inputs: torch.Tensor, rgb_stats: RGBType) -> torch.Tensor:
    min_val, max_val = normalized_bounds(rgb_stats, device=inputs.device, dtype=inputs.dtype)
    return torch.clamp(inputs, min=min_val, max=max_val)


def predict_labels(logits: torch.Tensor) -> torch.Tensor:
    return torch.argmax(logits, dim=1)


def validate_target(
    target: Optional[torch.Tensor], batch_size: int, num_classes: int, device: torch.device
) -> Optional[torch.Tensor]:
    if target is None:
        return None

    target = target.to(device=device, dtype=torch.long)
    if target.ndim == 0:
        target = target.view(1)

    if target.shape[0] != batch_size:
        raise ValueError(f"Target shape {target.shape[0]} must match batch size {batch_size}")

    if torch.any(target < 0) or torch.any(target >= num_classes):
        raise ValueError(f"Target values must be in range [0, {num_classes})")

    return target


def attack_success(
    logits: torch.Tensor,
    adv_logits: torch.Tensor,
    targeted: bool,
    target: Optional[torch.Tensor] = None,
    labels: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    adv_pred = predict_labels(adv_logits)
    if targeted is True:
        if target is None:
            raise ValueError("Target labels required for targeted attacks")

        return adv_pred.eq(target)

    base_labels = labels if labels is not None else predict_labels(logits)
    return adv_pred.ne(base_labels)



"""
RT-DETR v2 (Real-Time DEtection TRansformer), adapted from
https://github.com/lyuwenyu/RT-DETR/tree/main/rtdetrv2_pytorch

Paper "RT-DETRv2: Improved Baseline with Bag-of-Freebies for Real-Time Detection Transformer",
https://arxiv.org/abs/2407.17140
"""

# Reference license: Apache-2.0

import copy
import math
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import boxes as box_ops

from birder.common import training_utils
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.detection.base import DetectionBaseNet
from birder.net.detection.deformable_detr import HungarianMatcher
from birder.net.detection.deformable_detr import inverse_sigmoid
from birder.net.detection.rt_detr_v1 import HybridEncoder
from birder.net.detection.rt_detr_v1 import get_contrastive_denoising_training_group
from birder.net.detection.rt_detr_v1 import varifocal_loss
from birder.ops.msda import MultiScaleDeformableAttention as MSDA


def _get_clones(module: nn.Module, N: int) -> nn.ModuleList:
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


class MultiheadAttention(nn.Module):
    def __init__(self, d_model: int, num_heads: int, attn_drop: float = 0.0, proj_drop: float = 0.0) -> None:
        super().__init__()
        assert d_model % num_heads == 0, "d_model should be divisible by num_heads"

        self.num_heads = num_heads
        self.head_dim = d_model // num_heads
        self.scale = self.head_dim**-0.5

        self.q_proj = nn.Linear(d_model, d_model)
        self.k_proj = nn.Linear(d_model, d_model)
        self.v_proj = nn.Linear(d_model, d_model)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(d_model, d_model)
        self.proj_drop = nn.Dropout(proj_drop)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.xavier_uniform_(self.q_proj.weight)
        nn.init.xavier_uniform_(self.k_proj.weight)
        nn.init.xavier_uniform_(self.v_proj.weight)
        nn.init.xavier_uniform_(self.proj.weight)
        if self.q_proj.bias is not None:
            nn.init.zeros_(self.q_proj.bias)
            nn.init.zeros_(self.k_proj.bias)
            nn.init.zeros_(self.v_proj.bias)
            nn.init.zeros_(self.proj.bias)

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        attn_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        B, l_q, C = query.shape
        q = self.q_proj(query).reshape(B, l_q, self.num_heads, self.head_dim).transpose(1, 2)
        k = self.k_proj(key).reshape(B, key.size(1), self.num_heads, self.head_dim).transpose(1, 2)
        v = self.v_proj(value).reshape(B, value.size(1), self.num_heads, self.head_dim).transpose(1, 2)

        if attn_mask is not None:
            # attn_mask is (L, S) boolean where True = masked
            # SDPA expects True = attend, so we invert
            mask = ~attn_mask
        else:
            mask = None

        attn = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, attn_mask=mask, dropout_p=self.attn_drop.p if self.training else 0.0, scale=self.scale
        )

        attn = attn.transpose(1, 2).reshape(B, l_q, C)
        x = self.proj(attn)
        x = self.proj_drop(x)

        return x


class MultiScaleDeformableAttention(nn.Module):
    """
    Multi-Scale Deformable Attention with per-level point counts
    """

    def __init__(
        self,
        d_model: int,
        n_levels: int,
        n_heads: int,
        n_points: list[int],
        method: Literal["default", "discrete"] = "default",
        offset_scale: float = 0.5,
    ) -> None:
        super().__init__()
        assert d_model % n_heads == 0, "d_model must be divisible by n_heads"
        assert len(n_points) == n_levels, f"n_points list length must equal n_levels ({n_levels})"
        assert method in ("default", "discrete"), "method must be 'default' or 'discrete'"

        dim_per_head = d_model // n_heads
        if ((dim_per_head & (dim_per_head - 1) == 0) and dim_per_head != 0) is False:
            raise ValueError(
                "Set d_model in MultiScaleDeformableAttention to make the dimension of each attention head a power of 2"
            )

        self.im2col_step = 64
        self.d_model = d_model
        self.n_levels = n_levels
        self.n_heads = n_heads
        self.method = method
        self.offset_scale = offset_scale

        self.num_points = n_points
        num_points_scale = [1.0 / n for n in self.num_points for _ in range(n)]
        self.num_points_scale = nn.Buffer(torch.tensor(num_points_scale, dtype=torch.float32))
        self.total_points = sum(self.num_points)
        self.uniform_points = len(set(self.num_points)) == 1

        self.msda = MSDA()

        self.sampling_offsets = nn.Linear(d_model, n_heads * self.total_points * 2)
        self.attention_weights = nn.Linear(d_model, n_heads * self.total_points)
        self.value_proj = nn.Linear(d_model, d_model)
        self.output_proj = nn.Linear(d_model, d_model)

        self.reset_parameters()

        if method == "discrete":
            for param in self.sampling_offsets.parameters():
                param.requires_grad_(False)

    def reset_parameters(self) -> None:
        nn.init.zeros_(self.sampling_offsets.weight)
        thetas = torch.arange(self.n_heads, dtype=torch.float32) * (2.0 * math.pi / self.n_heads)
        grid_init = torch.stack([thetas.cos(), thetas.sin()], -1)
        grid_init = grid_init / grid_init.abs().max(-1, keepdim=True)[0]
        grid_init = grid_init.view(self.n_heads, 1, 2).repeat(1, self.total_points, 1)
        scaling = torch.concat([torch.arange(1, n + 1, dtype=torch.float32) for n in self.num_points]).view(1, -1, 1)
        grid_init = grid_init * scaling

        with torch.no_grad():
            self.sampling_offsets.bias = nn.Parameter(grid_init.view(-1))

        nn.init.zeros_(self.attention_weights.weight)
        nn.init.zeros_(self.attention_weights.bias)
        nn.init.xavier_uniform_(self.value_proj.weight)
        nn.init.zeros_(self.value_proj.bias)
        nn.init.xavier_uniform_(self.output_proj.weight)
        nn.init.zeros_(self.output_proj.bias)

    # pylint: disable=too-many-locals
    def forward(
        self,
        query: torch.Tensor,
        reference_points: torch.Tensor,
        input_flatten: torch.Tensor,
        input_spatial_shapes: torch.Tensor,
        src_shapes: list[list[int]],
        input_level_start_index: torch.Tensor,
        input_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        num_queries = query.size(1)
        N, sequence_length, _ = input_flatten.size()
        # assert (input_spatial_shapes[:, 0] * input_spatial_shapes[:, 1]).sum() == sequence_length

        value = self.value_proj(input_flatten)
        if input_padding_mask is not None:
            value = value.masked_fill(input_padding_mask[..., None], float(0))

        value = value.view(N, sequence_length, self.n_heads, self.d_model // self.n_heads)

        sampling_offsets = self.sampling_offsets(query).view(N, num_queries, self.n_heads, self.total_points, 2)
        attention_weights = self.attention_weights(query).view(N, num_queries, self.n_heads, self.total_points)
        attention_weights = F.softmax(attention_weights, dim=-1)

        if reference_points.shape[2] != self.n_levels:
            if reference_points.shape[2] == 1:
                reference_points = reference_points.expand(-1, -1, self.n_levels, -1)
            else:
                raise ValueError(
                    f"reference_points must have {self.n_levels} levels, but got {reference_points.shape[2]}"
                )

        if reference_points.shape[-1] == 2:
            offset_normalizer = torch.stack([input_spatial_shapes[..., 1], input_spatial_shapes[..., 0]], -1)
            sampling_locations_list = []
            offset_idx = 0
            for lvl in range(self.n_levels):
                n_pts = self.num_points[lvl]
                ref = reference_points[:, :, None, lvl : lvl + 1, :].expand(-1, -1, self.n_heads, n_pts, -1)
                off = sampling_offsets[:, :, :, offset_idx : offset_idx + n_pts, :]
                norm = offset_normalizer[lvl : lvl + 1].view(1, 1, 1, 1, 2)
                sampling_locations_list.append(ref + off / norm)
                offset_idx += n_pts

            sampling_locations = torch.concat(sampling_locations_list, dim=3)

        elif reference_points.shape[-1] == 4:
            sampling_locations_list = []
            offset_idx = 0
            num_points_scale = self.num_points_scale.to(dtype=query.dtype)
            for lvl in range(self.n_levels):
                n_pts = self.num_points[lvl]
                ref = reference_points[:, :, None, lvl : lvl + 1, :].expand(-1, -1, self.n_heads, n_pts, -1)
                off = sampling_offsets[:, :, :, offset_idx : offset_idx + n_pts, :]
                scale = num_points_scale[offset_idx : offset_idx + n_pts].view(1, 1, 1, n_pts, 1)
                sampling_locations_list.append(ref[..., :2] + off * scale * ref[..., 2:] * self.offset_scale)
                offset_idx += n_pts

            sampling_locations = torch.concat(sampling_locations_list, dim=3)

        else:
            raise ValueError(
                f"Last dim of reference_points must be 2 or 4, but get {reference_points.shape[-1]} instead"
            )

        if self.method == "discrete":
            output = self._forward_fallback(
                value, input_spatial_shapes, src_shapes, sampling_locations, attention_weights, method="discrete"
            )
        else:
            if self.uniform_points is True:
                n_pts = self.num_points[0]
                sampling_locations = sampling_locations.view(N, num_queries, self.n_heads, self.n_levels, n_pts, 2)
                attention_weights = attention_weights.view(N, num_queries, self.n_heads, self.n_levels, n_pts)
                output = self.msda(
                    value,
                    input_spatial_shapes,
                    input_level_start_index,
                    sampling_locations,
                    attention_weights,
                    self.im2col_step,
                    src_shapes,
                )
            else:
                output = self._forward_fallback(
                    value, input_spatial_shapes, src_shapes, sampling_locations, attention_weights, method="default"
                )

        output = self.output_proj(output)
        return output

    def _forward_fallback(
        self,
        value: torch.Tensor,
        spatial_shapes: torch.Tensor,
        src_shapes: list[list[int]],
        sampling_locations: torch.Tensor,
        attention_weights: torch.Tensor,
        method: str = "default",
    ) -> torch.Tensor:
        B, _, n_heads, head_dim = value.size()
        num_queries = sampling_locations.size(1)

        sampling_grids = 2 * sampling_locations - 1
        split_shape: list[int] = (spatial_shapes[:, 0] * spatial_shapes[:, 1]).tolist()
        value_list = value.permute(0, 2, 3, 1).flatten(0, 1).split(split_shape, dim=-1)
        sampling_grids = sampling_grids.permute(0, 2, 1, 3, 4).flatten(0, 1)
        sampling_locations_list = sampling_grids.split(self.num_points, dim=-2)

        sampling_value_list = []
        for level, (H, W) in enumerate(src_shapes):
            value_l = value_list[level].reshape(B * n_heads, head_dim, H, W)
            sampling_grid_l = sampling_locations_list[level]

            if method == "default":
                sampling_value_l = F.grid_sample(
                    value_l,
                    sampling_grid_l,
                    mode="bilinear",
                    padding_mode="zeros",
                    align_corners=False,
                )
            else:
                sampling_grid_l = sampling_grid_l.clone()
                sampling_grid_l[..., 0] += 1.0 / W
                sampling_grid_l[..., 1] += 1.0 / H
                sampling_value_l = F.grid_sample(
                    value_l,
                    sampling_grid_l,
                    mode="nearest",
                    padding_mode="border",
                    align_corners=False,
                )

                # Original upstream code (expected grid of [0, 1])
                # e.g. without the 'sampling_grids = 2 * sampling_locations - 1'
                #
                # n_pts = self.num_points[level]
                # sampling_coord = (sampling_grid_l * torch.tensor([[W, H]], device=value.device) + 0.5).to(torch.int64)
                # sampling_coord[..., 0] = sampling_coord[..., 0].clamp(0, W - 1)
                # sampling_coord[..., 1] = sampling_coord[..., 1].clamp(0, H - 1)
                # sampling_coord = sampling_coord.reshape(B * n_heads, num_queries * n_pts, 2)
                # s_idx = (
                #     torch.arange(sampling_coord.shape[0], device=value.device)
                #     .unsqueeze(-1)
                #     .repeat(1, sampling_coord.shape[1])
                # )
                # sampling_value_l = value_l[s_idx, :, sampling_coord[..., 1], sampling_coord[..., 0]]
                # ... = sampling_value_l.permute(0, 2, 1).reshape(B * n_heads, head_dim, num_queries, n_pts)

            sampling_value_list.append(sampling_value_l)

        attn_weights = attention_weights.permute(0, 2, 1, 3).reshape(B * n_heads, 1, num_queries, sum(self.num_points))
        output = torch.concat(sampling_value_list, dim=-1) * attn_weights
        output = output.sum(-1).reshape(B, n_heads * head_dim, num_queries)

        return output.permute(0, 2, 1)


class TransformerDecoderLayer(nn.Module):
    def __init__(
        self,
        d_model: int,
        d_ffn: int,
        dropout: float,
        n_levels: int,
        n_heads: int,
        n_points: list[int],
        method: Literal["default", "discrete"] = "default",
        offset_scale: float = 0.5,
    ) -> None:
        super().__init__()

        # Self attention
        self.self_attn = MultiheadAttention(d_model, n_heads, attn_drop=dropout)
        self.norm1 = nn.LayerNorm(d_model)

        # Cross attention
        self.cross_attn = MultiScaleDeformableAttention(
            d_model, n_levels, n_heads, n_points, method=method, offset_scale=offset_scale
        )
        self.norm2 = nn.LayerNorm(d_model)

        # FFN
        self.linear1 = nn.Linear(d_model, d_ffn)
        self.linear2 = nn.Linear(d_ffn, d_model)
        self.norm3 = nn.LayerNorm(d_model)

        self.activation = nn.ReLU()
        self.dropout = nn.Dropout(dropout)

    def forward(
        self,
        tgt: torch.Tensor,
        query_pos: torch.Tensor,
        reference_points: torch.Tensor,
        src: torch.Tensor,
        src_spatial_shapes: torch.Tensor,
        src_shapes: list[list[int]],
        level_start_index: torch.Tensor,
        src_padding_mask: Optional[torch.Tensor],
        self_attn_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        # Self attention
        q_k = tgt + query_pos

        tgt2 = self.self_attn(q_k, q_k, tgt, attn_mask=self_attn_mask)
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm1(tgt)

        # Cross attention
        tgt2 = self.cross_attn(
            tgt + query_pos, reference_points, src, src_spatial_shapes, src_shapes, level_start_index, src_padding_mask
        )
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm2(tgt)

        # FFN
        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt))))
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm3(tgt)

        return tgt


# pylint: disable=invalid-name
class RT_DETRDecoder(nn.Module):
    """
    RT-DETR v2 Decoder with top-k query selection
    """

    def __init__(
        self,
        hidden_dim: int,
        num_classes: int,
        num_queries: int,
        num_decoder_layers: int,
        num_levels: int,
        num_heads: int,
        dim_feedforward: int,
        dropout: float,
        num_decoder_points: list[int],
        method: Literal["default", "discrete"] = "default",
        offset_scale: float = 0.5,
    ) -> None:
        super().__init__()
        self.hidden_dim = hidden_dim
        self.num_queries = num_queries
        self.num_levels = num_levels

        self.enc_output = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.LayerNorm(hidden_dim),
        )
        self.enc_score_head = nn.Linear(hidden_dim, num_classes)
        self.enc_bbox_head = MLP(hidden_dim, [hidden_dim, hidden_dim, 4], activation_layer=nn.ReLU)

        decoder_layer = TransformerDecoderLayer(
            hidden_dim,
            dim_feedforward,
            dropout,
            num_levels,
            num_heads,
            num_decoder_points,
            method=method,
            offset_scale=offset_scale,
        )
        self.layers = _get_clones(decoder_layer, num_decoder_layers)

        self.query_pos_head = MLP(4, [2 * hidden_dim, hidden_dim], activation_layer=nn.ReLU)
        self.class_embed = nn.ModuleList([nn.Linear(hidden_dim, num_classes) for _ in range(num_decoder_layers)])
        self.bbox_embed = nn.ModuleList(
            [MLP(hidden_dim, [hidden_dim, hidden_dim, 4], activation_layer=nn.ReLU) for _ in range(num_decoder_layers)]
        )
        self.use_cache = True
        self._anchor_cache: dict[str, tuple[torch.Tensor, torch.Tensor]] = {}

        # Weights initialization
        prior_prob = 0.01
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        nn.init.xavier_uniform_(self.enc_output[0].weight)
        nn.init.xavier_uniform_(self.enc_score_head.weight)
        nn.init.constant_(self.enc_score_head.bias, bias_value)
        nn.init.zeros_(self.enc_bbox_head[-2].weight)
        nn.init.zeros_(self.enc_bbox_head[-2].bias)
        for class_embed in self.class_embed:
            nn.init.constant_(class_embed.bias, bias_value)

        for bbox_embed in self.bbox_embed:
            nn.init.zeros_(bbox_embed[-2].weight)
            nn.init.zeros_(bbox_embed[-2].bias)

    def set_cache_enabled(self, enabled: bool) -> None:
        self.use_cache = enabled
        if enabled is False:
            self.clear_cache()

    def clear_cache(self) -> None:
        self._anchor_cache.clear()

    def _generate_anchors(
        self,
        spatial_shapes: list[list[int]],
        grid_size: float = 0.05,
        device: torch.device = torch.device("cpu"),
        dtype: torch.dtype = torch.float32,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        cache_key: Optional[str] = None
        use_cache = self.use_cache is True and torch.jit.is_tracing() is False and torch.jit.is_scripting() is False
        if use_cache is True:
            spatial_key = ",".join(f"{int(h)}x{int(w)}" for h, w in spatial_shapes)
            cache_key = f"{spatial_key}_{grid_size}_{device}_{dtype}"
            cached = self._anchor_cache.get(cache_key)
            if cached is not None:
                return cached

        anchors = []
        for lvl, (h, w) in enumerate(spatial_shapes):
            grid_y, grid_x = torch.meshgrid(
                torch.arange(h, dtype=dtype, device=device),
                torch.arange(w, dtype=dtype, device=device),
                indexing="ij",
            )
            grid_xy = torch.stack([grid_x, grid_y], dim=-1)
            valid_wh = torch.tensor([w, h], dtype=dtype, device=device)
            grid_xy = (grid_xy.unsqueeze(0) + 0.5) / valid_wh
            wh = torch.ones_like(grid_xy) * grid_size * (2.0**lvl)
            anchors.append(torch.concat([grid_xy, wh], dim=-1).reshape(-1, h * w, 4))

        anchors = torch.concat(anchors, dim=1)
        eps = 0.01
        valid_mask = ((anchors > eps) * (anchors < 1 - eps)).all(dim=-1, keepdim=True)
        anchors = torch.log(anchors / (1 - anchors))
        anchors = torch.where(valid_mask, anchors, torch.inf)

        if cache_key is not None:
            self._anchor_cache[cache_key] = (anchors, valid_mask)

        return (anchors, valid_mask)

    def _get_decoder_input(
        self,
        memory: torch.Tensor,
        spatial_shapes: list[list[int]],
        memory_padding_mask: Optional[torch.Tensor] = None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        anchors, valid_mask = self._generate_anchors(spatial_shapes, device=memory.device, dtype=memory.dtype)
        if memory_padding_mask is not None:
            valid_mask = valid_mask & ~memory_padding_mask.unsqueeze(-1)

        memory = valid_mask.to(memory.dtype) * memory
        output_memory = self.enc_output(memory)
        enc_outputs_class = self.enc_score_head(output_memory)
        if memory_padding_mask is not None:
            enc_outputs_class = enc_outputs_class.masked_fill(memory_padding_mask[..., None], float("-inf"))

        enc_outputs_coord_unact = self.enc_bbox_head(output_memory) + anchors

        # Select top-k queries based on classification confidence
        _, topk_ind = torch.topk(enc_outputs_class.max(dim=-1).values, self.num_queries, dim=1)

        # Gather reference points
        reference_points_unact = enc_outputs_coord_unact.gather(
            dim=1, index=topk_ind.unsqueeze(-1).expand(-1, -1, enc_outputs_coord_unact.shape[-1])
        )

        enc_topk_bboxes = reference_points_unact.sigmoid()

        # Gather encoder logits for loss computation
        enc_topk_logits = enc_outputs_class.gather(
            dim=1, index=topk_ind.unsqueeze(-1).expand(-1, -1, enc_outputs_class.shape[-1])
        )

        # Extract region features
        target = output_memory.gather(dim=1, index=topk_ind.unsqueeze(-1).expand(-1, -1, output_memory.shape[-1]))
        target = target.detach()

        return (target, reference_points_unact.detach(), enc_topk_bboxes, enc_topk_logits)

    def forward(  # pylint: disable=too-many-locals
        self,
        feats: list[torch.Tensor],
        spatial_shapes: list[list[int]],
        level_start_index: list[int],
        denoising_class: Optional[torch.Tensor] = None,
        denoising_bbox_unact: Optional[torch.Tensor] = None,
        attn_mask: Optional[torch.Tensor] = None,
        padding_mask: Optional[list[torch.Tensor]] = None,
        return_intermediates: bool = True,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        memory = []
        mask_flatten = []
        for idx, feat in enumerate(feats):
            feat_flat = feat.flatten(2).permute(0, 2, 1)  # (B, H*W, C)
            memory.append(feat_flat)
            if padding_mask is not None:
                mask_flatten.append(padding_mask[idx].flatten(1))

        memory = torch.concat(memory, dim=1)
        memory_padding_mask = torch.concat(mask_flatten, dim=1) if mask_flatten else None

        # Get decoder input (query selection)
        target, init_ref_points_unact, enc_topk_bboxes, enc_topk_logits = self._get_decoder_input(
            memory, spatial_shapes, memory_padding_mask
        )

        # Concatenate denoising queries if provided
        if denoising_class is not None and denoising_bbox_unact is not None:
            target = torch.concat([denoising_class, target], dim=1)
            init_ref_points_unact = torch.concat([denoising_bbox_unact, init_ref_points_unact], dim=1)

        # Prepare spatial shapes and level start index as tensors
        spatial_shapes_tensor = torch.tensor(spatial_shapes, dtype=torch.long, device=memory.device)
        level_start_index_tensor = torch.tensor(level_start_index, dtype=torch.long, device=memory.device)

        # Decoder forward
        bboxes_list: list[torch.Tensor] = []
        logits_list: list[torch.Tensor] = []
        reference_points = init_ref_points_unact.sigmoid()
        for decoder_layer, bbox_head, class_head in zip(self.layers, self.bbox_embed, self.class_embed):
            query_pos = self.query_pos_head(reference_points)
            reference_points_input = reference_points.unsqueeze(2).expand(-1, -1, len(spatial_shapes), -1)
            target = decoder_layer(
                target,
                query_pos,
                reference_points_input,
                memory,
                spatial_shapes_tensor,
                spatial_shapes,
                level_start_index_tensor,
                memory_padding_mask,
                attn_mask,
            )

            bbox_delta = bbox_head(target)
            new_reference_points = inverse_sigmoid(reference_points) + bbox_delta
            new_reference_points = new_reference_points.sigmoid()

            # Classification
            class_logits = class_head(target)

            if return_intermediates is True:
                bboxes_list.append(new_reference_points)
                logits_list.append(class_logits)

            # Update reference points for next layer
            reference_points = new_reference_points.detach()

        if return_intermediates is True:
            out_bboxes = torch.stack(bboxes_list)
            out_logits = torch.stack(logits_list)
        else:
            out_bboxes = new_reference_points
            out_logits = class_logits

        return (out_bboxes, out_logits, enc_topk_bboxes, enc_topk_logits)


# pylint: disable=invalid-name
class RT_DETR_v2(DetectionBaseNet):
    default_size = (640, 640)

    def __init__(
        self,
        num_classes: int,
        backbone: DetectorBackbone,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
        export_mode: bool = False,
    ) -> None:
        super().__init__(num_classes, backbone, config=config, size=size, export_mode=export_mode)
        assert self.config is not None, "must set config"

        self.reparameterized = False

        # Sigmoid based classification (no background class in predictions)
        self.num_classes = self.num_classes - 1

        hidden_dim = self.config.get("hidden_dim", 256)
        num_heads = self.config.get("num_heads", 8)
        dim_feedforward = self.config.get("dim_feedforward", 1024)
        dropout: float = self.config.get("dropout", 0.0)
        num_encoder_layers: int = self.config.get("num_encoder_layers", 1)
        num_decoder_layers: int = self.config["num_decoder_layers"]
        num_queries: int = self.config.get("num_queries", 300)
        expansion: float = self.config.get("expansion", 1.0)
        depth_multiplier: float = self.config.get("depth_multiplier", 1.0)
        use_giou: bool = self.config.get("use_giou", True)
        num_denoising: int = self.config.get("num_denoising", 100)
        label_noise_ratio: float = self.config.get("label_noise_ratio", 0.5)
        box_noise_scale: float = self.config.get("box_noise_scale", 1.0)
        num_decoder_points: list[int] = self.config.get("num_decoder_points", [4, 4, 4])
        method: Literal["default", "discrete"] = self.config.get("method", "default")
        offset_scale: float = self.config.get("offset_scale", 0.5)

        self.hidden_dim = hidden_dim
        self.num_queries = num_queries
        self.num_denoising = num_denoising
        self.label_noise_ratio = label_noise_ratio
        self.box_noise_scale = box_noise_scale

        self.backbone.return_channels = self.backbone.return_channels[-3:]
        self.backbone.return_stages = self.backbone.return_stages[-3:]
        self.num_levels = len(self.backbone.return_channels)

        self.encoder = HybridEncoder(
            in_channels=self.backbone.return_channels,
            hidden_dim=hidden_dim,
            num_encoder_layers=num_encoder_layers,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            num_heads=num_heads,
            expansion=expansion,
            depth_multiplier=depth_multiplier,
        )
        self.decoder = RT_DETRDecoder(
            hidden_dim=hidden_dim,
            num_classes=self.num_classes,
            num_queries=self.num_queries,
            num_decoder_layers=num_decoder_layers,
            num_levels=self.num_levels,
            num_heads=num_heads,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            num_decoder_points=num_decoder_points,
            method=method,
            offset_scale=offset_scale,
        )

        self.matcher = HungarianMatcher(cost_class=2.0, cost_bbox=5.0, cost_giou=2.0, use_giou=use_giou)

        # Denoising class embedding for Contrastive denoising (CDN) training
        if self.num_denoising > 0:
            self.denoising_class_embed = nn.Embedding(self.num_classes + 1, hidden_dim, padding_idx=self.num_classes)

        if self.export_mode is False:
            self.forward = torch.compiler.disable(recursive=False)(self.forward)  # type: ignore[method-assign]

    def _set_cache_enabled(self, enabled: bool) -> None:
        self.encoder.set_cache_enabled(enabled)
        self.decoder.set_cache_enabled(enabled)

    def clear_cache(self) -> None:
        self.encoder.clear_cache()
        self.decoder.clear_cache()

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        super().adjust_size(new_size)
        self.clear_cache()

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        super().set_dynamic_size(dynamic_size)
        self._set_cache_enabled(dynamic_size is False)

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes

        self.decoder.enc_score_head = nn.Linear(self.hidden_dim, num_classes)
        self.decoder.class_embed = nn.ModuleList(
            [nn.Linear(self.hidden_dim, num_classes) for _ in range(len(self.decoder.layers))]
        )

        if self.num_denoising > 0:
            self.denoising_class_embed = nn.Embedding(num_classes + 1, self.hidden_dim, padding_idx=num_classes)

        prior_prob = 0.01
        bias_value = -math.log((1 - prior_prob) / prior_prob)
        nn.init.constant_(self.decoder.enc_score_head.bias, bias_value)
        for class_embed in self.decoder.class_embed:
            nn.init.constant_(class_embed.bias, bias_value)

    def freeze(self, freeze_classifier: bool = True) -> None:
        for param in self.parameters():
            param.requires_grad_(False)

        if freeze_classifier is False:
            for param in self.decoder.class_embed.parameters():
                param.requires_grad_(True)
            for param in self.decoder.enc_score_head.parameters():
                param.requires_grad_(True)
            if self.num_denoising > 0:
                for param in self.denoising_class_embed.parameters():
                    param.requires_grad_(True)

    @staticmethod
    def _get_src_permutation_idx(indices: list[tuple[torch.Tensor, torch.Tensor]]) -> tuple[torch.Tensor, torch.Tensor]:
        batch_idx = torch.concat([torch.full_like(src, i) for i, (src, _) in enumerate(indices)])
        src_idx = torch.concat([src for (src, _) in indices])
        return (batch_idx, src_idx)

    def _compute_layer_losses(
        self,
        cls_logits: torch.Tensor,
        box_output: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        indices: list[tuple[torch.Tensor, torch.Tensor]],
        num_boxes: float,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        idx = self._get_src_permutation_idx(indices)

        src_boxes = box_output[idx]
        target_boxes = torch.concat([t["boxes"][i] for t, (_, i) in zip(targets, indices)], dim=0)

        src_boxes_xyxy = box_ops.box_convert(src_boxes, in_fmt="cxcywh", out_fmt="xyxy")
        target_boxes_xyxy = box_ops.box_convert(target_boxes, in_fmt="cxcywh", out_fmt="xyxy")

        # IoU for varifocal loss (class loss)
        ious = torch.diag(box_ops.box_iou(src_boxes_xyxy, target_boxes_xyxy)).detach()

        # Classification loss
        target_classes_o = torch.concat([t["labels"][J] for t, (_, J) in zip(targets, indices)], dim=0)
        target_classes = torch.full(cls_logits.shape[:2], self.num_classes, dtype=torch.int64, device=cls_logits.device)
        target_classes[idx] = target_classes_o

        target_classes_onehot = torch.zeros(
            [cls_logits.shape[0], cls_logits.shape[1], cls_logits.shape[2] + 1],
            dtype=cls_logits.dtype,
            layout=cls_logits.layout,
            device=cls_logits.device,
        )
        target_classes_onehot.scatter_(2, target_classes.unsqueeze(-1), 1)
        target_classes_onehot = target_classes_onehot[:, :, :-1]

        target_score_o = torch.zeros(cls_logits.shape[:2], dtype=cls_logits.dtype, device=cls_logits.device)
        target_score_o[idx] = ious.to(cls_logits.dtype)
        target_score = target_score_o.unsqueeze(-1) * target_classes_onehot

        loss = varifocal_loss(cls_logits, target_score, target_classes_onehot, alpha=0.75, gamma=2.0)
        loss_ce = (loss.mean(1).sum() / num_boxes) * cls_logits.shape[1]

        # Box L1 loss
        loss_bbox = F.l1_loss(src_boxes, target_boxes, reduction="none").sum() / num_boxes

        # GIoU loss
        loss_giou = (1 - torch.diag(box_ops.generalized_box_iou(src_boxes_xyxy, target_boxes_xyxy))).sum() / num_boxes

        return (loss_ce, loss_bbox, loss_giou)

    def _compute_denoising_loss(
        self,
        dn_out_bboxes: torch.Tensor,
        dn_out_logits: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        dn_meta: dict[str, Any],
        num_boxes: float,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        dn_positive_idx = dn_meta["dn_positive_idx"]
        num_groups = dn_meta["dn_num_group"]

        loss_ce_list = []
        loss_bbox_list = []
        loss_giou_list = []

        dn_num_boxes = max(num_boxes * num_groups, 1.0)
        for layer_idx in range(dn_out_logits.shape[0]):
            # Construct indices from denoising metadata
            indices = []
            for batch_idx, pos_idx in enumerate(dn_positive_idx):
                if len(pos_idx) > 0:
                    src_idx = pos_idx
                    num_gt = len(targets[batch_idx]["labels"])
                    tgt_idx = torch.arange(num_gt, device=pos_idx.device).repeat(num_groups)
                    indices.append((src_idx, tgt_idx))
                else:
                    indices.append(
                        (
                            torch.tensor([], dtype=torch.long, device=dn_out_logits.device),
                            torch.tensor([], dtype=torch.long, device=dn_out_logits.device),
                        )
                    )

            loss_ce, loss_bbox, loss_giou = self._compute_layer_losses(
                dn_out_logits[layer_idx], dn_out_bboxes[layer_idx], targets, indices, dn_num_boxes
            )
            loss_ce_list.append(loss_ce)
            loss_bbox_list.append(loss_bbox)
            loss_giou_list.append(loss_giou)

        loss_ce_dn = torch.stack(loss_ce_list).sum()
        loss_bbox_dn = torch.stack(loss_bbox_list).sum()
        loss_giou_dn = torch.stack(loss_giou_list).sum()

        return (loss_ce_dn, loss_bbox_dn, loss_giou_dn)

    def _compute_loss_from_outputs(
        self,
        targets: list[dict[str, torch.Tensor]],
        out_bboxes: torch.Tensor,
        out_logits: torch.Tensor,
        enc_topk_bboxes: torch.Tensor,
        enc_topk_logits: torch.Tensor,
        dn_out_bboxes: Optional[torch.Tensor] = None,
        dn_out_logits: Optional[torch.Tensor] = None,
        dn_meta: Optional[dict[str, Any]] = None,
    ) -> dict[str, torch.Tensor]:
        # Compute the average number of target boxes across all nodes
        num_boxes = sum(len(t["labels"]) for t in targets)
        num_boxes = torch.as_tensor([num_boxes], dtype=torch.float, device=out_logits.device)
        if training_utils.is_dist_available_and_initialized() is True:
            torch.distributed.all_reduce(num_boxes)

        num_boxes = torch.clamp(num_boxes / training_utils.get_world_size(), min=1)

        loss_ce_list = []
        loss_bbox_list = []
        loss_giou_list = []

        # Decoder losses (all layers)
        for layer_idx in range(out_logits.shape[0]):
            indices = self.matcher(out_logits[layer_idx], out_bboxes[layer_idx], targets)
            loss_ce, loss_bbox, loss_giou = self._compute_layer_losses(
                out_logits[layer_idx], out_bboxes[layer_idx], targets, indices, num_boxes
            )
            loss_ce_list.append(loss_ce)
            loss_bbox_list.append(loss_bbox)
            loss_giou_list.append(loss_giou)

        # Encoder auxiliary loss
        enc_indices = self.matcher(enc_topk_logits, enc_topk_bboxes, targets)
        loss_ce, loss_bbox, loss_giou = self._compute_layer_losses(
            enc_topk_logits, enc_topk_bboxes, targets, enc_indices, num_boxes
        )
        loss_ce_list.append(loss_ce)
        loss_bbox_list.append(loss_bbox)
        loss_giou_list.append(loss_giou)

        loss_ce = torch.stack(loss_ce_list).sum()  # VFL weight is 1
        loss_bbox = torch.stack(loss_bbox_list).sum() * 5
        loss_giou = torch.stack(loss_giou_list).sum() * 2

        # Add denoising loss if available
        if dn_out_bboxes is not None and dn_out_logits is not None and dn_meta is not None:
            loss_ce_dn, loss_bbox_dn, loss_giou_dn = self._compute_denoising_loss(
                dn_out_bboxes, dn_out_logits, targets, dn_meta, num_boxes
            )
            loss_ce = loss_ce + loss_ce_dn
            loss_bbox = loss_bbox + loss_bbox_dn * 5
            loss_giou = loss_giou + loss_giou_dn * 2

        losses = {
            "labels": loss_ce,
            "boxes": loss_bbox,
            "giou": loss_giou,
        }

        return losses

    @torch.jit.unused  # type: ignore[untyped-decorator]
    @torch.compiler.disable()  # type: ignore[untyped-decorator]
    def compute_loss(
        self,
        encoder_features: list[torch.Tensor],
        spatial_shapes: list[list[int]],
        level_start_index: list[int],
        targets: list[dict[str, torch.Tensor]],
        images: Any,
        masks: Optional[list[torch.Tensor]] = None,
    ) -> dict[str, torch.Tensor]:
        for idx, target in enumerate(targets):
            boxes = target["boxes"]
            boxes = box_ops.box_convert(boxes, in_fmt="xyxy", out_fmt="cxcywh")
            scale = images.image_sizes[idx].flip(0).repeat(2).float()  # flip to [W, H], repeat to [W, H, W, H]
            boxes = boxes / scale
            targets[idx]["boxes"] = boxes
            targets[idx]["labels"] = target["labels"] - 1  # No background

        denoising_class, denoising_bbox_unact, attn_mask, dn_meta = self._prepare_cdn_queries(targets)

        out_bboxes, out_logits, enc_topk_bboxes, enc_topk_logits = self.decoder(
            encoder_features,
            spatial_shapes,
            level_start_index,
            denoising_class,
            denoising_bbox_unact,
            attn_mask,
            masks,
        )

        if dn_meta is not None:
            dn_num_split, _num_queries = dn_meta["dn_num_split"]
            dn_out_bboxes = out_bboxes[:, :, :dn_num_split]
            dn_out_logits = out_logits[:, :, :dn_num_split]
            out_bboxes = out_bboxes[:, :, dn_num_split:]
            out_logits = out_logits[:, :, dn_num_split:]
        else:
            dn_out_bboxes = None
            dn_out_logits = None

        losses: dict[str, torch.Tensor] = self._compute_loss_from_outputs(
            targets, out_bboxes, out_logits, enc_topk_bboxes, enc_topk_logits, dn_out_bboxes, dn_out_logits, dn_meta
        )

        return losses

    def postprocess_detections(
        self, class_logits: torch.Tensor, box_regression: torch.Tensor, image_sizes: torch.Tensor
    ) -> list[dict[str, torch.Tensor]]:
        prob = class_logits.sigmoid()
        topk_values, topk_indexes = torch.topk(prob.view(class_logits.shape[0], -1), k=self.decoder.num_queries, dim=1)
        scores = topk_values
        topk_boxes = topk_indexes // class_logits.shape[2]
        labels = topk_indexes % class_logits.shape[2]
        labels += 1  # Background offset

        # Convert to [x0, y0, x1, y1] format
        boxes = box_ops.box_convert(box_regression, in_fmt="cxcywh", out_fmt="xyxy")
        boxes = torch.gather(boxes, 1, topk_boxes.unsqueeze(-1).expand(-1, -1, 4))

        # Convert from relative [0, 1] to absolute [0, height] coordinates
        img_h, img_w = image_sizes.unbind(1)
        scale_fct = torch.stack([img_w, img_h, img_w, img_h], dim=1)
        boxes = boxes * scale_fct[:, None, :]

        detections: list[dict[str, torch.Tensor]] = []
        for s, l, b in zip(scores, labels, boxes):
            detections.append(
                {
                    "boxes": b,
                    "scores": s,
                    "labels": l,
                }
            )

        return detections

    @torch.jit.unused  # type: ignore[untyped-decorator]
    def _prepare_cdn_queries(
        self, targets: list[dict[str, torch.Tensor]]
    ) -> tuple[Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor], Optional[dict[str, Any]]]:
        if self.num_denoising > 0:
            result: tuple[
                Optional[torch.Tensor], Optional[torch.Tensor], Optional[torch.Tensor], Optional[dict[str, Any]]
            ] = get_contrastive_denoising_training_group(
                targets,
                self.num_classes,
                self.num_queries,
                self.denoising_class_embed,
                num_denoising_queries=self.num_denoising,
                label_noise_ratio=self.label_noise_ratio,
                box_noise_scale=self.box_noise_scale,
            )
            return result

        return (None, None, None, None)

    def forward_net(
        self, x: torch.Tensor, masks: Optional[torch.Tensor]
    ) -> tuple[list[torch.Tensor], Optional[list[torch.Tensor]]]:
        features: dict[str, torch.Tensor] = self.backbone.detection_features(x)
        feature_list = list(features.values())

        mask_list: Optional[list[torch.Tensor]] = None
        if masks is not None:
            mask_list = []
            for feat in feature_list:
                m = F.interpolate(masks[None].float(), size=feat.shape[-2:], mode="nearest").to(torch.bool)[0]
                mask_list.append(m)

        encoder_features = self.encoder(feature_list, masks=mask_list)

        return (encoder_features, mask_list)

    def forward(
        self,
        x: torch.Tensor,
        targets: Optional[list[dict[str, torch.Tensor]]] = None,
        masks: Optional[torch.Tensor] = None,
        image_sizes: Optional[list[tuple[int, int]]] = None,
    ) -> tuple[list[dict[str, torch.Tensor]], dict[str, torch.Tensor]]:
        self._input_check(targets)
        images = self._to_img_list(x, image_sizes)

        encoder_features, mask_list = self.forward_net(x, masks)

        # Prepare spatial shapes and level start index
        spatial_shapes: list[list[int]] = []
        level_start_index: list[int] = [0]
        for feat in encoder_features:
            H = feat.shape[2]
            W = feat.shape[3]
            spatial_shapes.append([H, W])
            level_start_index.append(H * W + level_start_index[-1])

        level_start_index.pop()

        detections: list[dict[str, torch.Tensor]] = []
        losses: dict[str, torch.Tensor] = {}
        if self.training is True:
            assert targets is not None, "targets should not be None when in training mode"
            losses = self.compute_loss(encoder_features, spatial_shapes, level_start_index, targets, images, mask_list)
        else:
            # Inference path - no CDN
            out_bboxes, out_logits, _, _ = self.decoder(
                encoder_features, spatial_shapes, level_start_index, padding_mask=mask_list, return_intermediates=False
            )
            detections = self.postprocess_detections(out_logits, out_bboxes, images.image_sizes)

        return (detections, losses)

    @torch.no_grad()  # type: ignore[untyped-decorator]
    def reparameterize_model(self) -> None:
        if self.reparameterized is True:
            return

        for module in self.modules():
            if hasattr(module, "reparameterize") is True:
                module.reparameterize()

        self.reparameterized = True


registry.register_model_config(
    "rt_detr_v2_s",
    RT_DETR_v2,
    config={
        "num_decoder_layers": 3,
        "expansion": 0.5,
    },
)
registry.register_model_config(
    "rt_detr_v2_s_dsp",
    RT_DETR_v2,
    config={
        "num_decoder_layers": 3,
        "expansion": 0.5,
        "method": "discrete",
    },
)
registry.register_model_config(
    "rt_detr_v2",
    RT_DETR_v2,
    config={
        "num_decoder_layers": 6,
    },
)
registry.register_model_config(
    "rt_detr_v2_dsp",
    RT_DETR_v2,
    config={
        "num_decoder_layers": 6,
        "method": "discrete",
    },
)
registry.register_model_config(
    "rt_detr_v2_l",
    RT_DETR_v2,
    config={
        "num_decoder_layers": 6,
        "expansion": 1.0,
        "depth_multiplier": 1.0,
        "num_heads": 12,  # Deviates from upstream to keep head_dim=32 (power of 2) for MSDA kernel
        "hidden_dim": 384,
        "dim_feedforward": 2048,
    },
)
registry.register_model_config(
    "rt_detr_v2_l_dsp",
    RT_DETR_v2,
    config={
        "num_decoder_layers": 6,
        "expansion": 1.0,
        "depth_multiplier": 1.0,
        "num_heads": 12,  # Deviates from upstream to keep head_dim=32 (power of 2) for MSDA kernel
        "hidden_dim": 384,
        "dim_feedforward": 2048,
        "method": "discrete",
    },
)



"""
AMI clustering for unsupervised embedding evaluation

Paper "An Empirical Study into Clustering of Unseen Datasets with Self-Supervised Encoders",
https://arxiv.org/abs/2406.02465
"""

import logging
from typing import Optional

import numpy as np
import numpy.typing as npt
from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics import adjusted_mutual_info_score

from birder.eval._embeddings import l2_normalize

logger = logging.getLogger(__name__)

try:
    import umap

    _HAS_UMAP = True
except ImportError:
    _HAS_UMAP = False


def evaluate_ami(
    features: npt.NDArray[np.float32],
    labels: npt.NDArray[np.int_],
    n_clusters: int,
    umap_dim: int = 50,
    l2_normalize_features: bool = True,
    seed: Optional[int] = None,
) -> float:
    """
    Evaluate embedding quality using UMAP + Agglomerative Clustering + AMI

    Parameters
    ----------
    features
        Feature array of shape (n_samples, embedding_dim).
    labels
        True labels of shape (n_samples,).
    n_clusters
        Number of clusters (should match number of true classes).
    umap_dim
        Target dimensionality for UMAP reduction.
    l2_normalize_features
        If True, applies row-wise L2 normalization before UMAP.
    seed
        Random seed for UMAP reproducibility. When None, uses all available
        cores (n_jobs=-1) but results are non-deterministic. When set, forces n_jobs=1.

    Returns
    -------
    ami_score
        Adjusted Mutual Information score between true labels and cluster assignments.
    """

    assert _HAS_UMAP, "'pip install umap-learn' to use AMI evaluation"

    if seed is None:
        logger.warning("No seed set, UMAP results will be non-deterministic (using n_jobs=-1)")
        n_jobs = -1
    else:
        n_jobs = 1

    if l2_normalize_features is True:
        features = l2_normalize(features)

    reducer = umap.UMAP(n_components=umap_dim, min_dist=0.0, n_jobs=n_jobs, random_state=seed)
    features_reduced = reducer.fit_transform(features)

    clustering = AgglomerativeClustering(n_clusters=n_clusters, linkage="ward")
    cluster_assignments = clustering.fit_predict(features_reduced)

    return float(adjusted_mutual_info_score(labels, cluster_assignments))



import logging
import typing
import unittest
from typing import Any

import torch
from PIL import Image
from torchvision import tv_tensors
from torchvision.transforms import v2

from birder.data.transforms import classification
from birder.data.transforms import detection
from birder.data.transforms import mosaic

logging.disable(logging.CRITICAL)


class TestTransforms(unittest.TestCase):
    def test_classification(self) -> None:
        # Get rgb
        for rgb_mode in typing.get_args(classification.RGBMode):
            rgb_stats = classification.get_rgb_stats(rgb_mode)
            self.assertIsInstance(rgb_stats, dict)
            self.assertIn("mean", rgb_stats)
            self.assertIn("std", rgb_stats)
            self.assertEqual(len(rgb_stats["mean"]), 3)
            self.assertEqual(len(rgb_stats["std"]), 3)

        # Get mixup / cutmix
        mixup_cutmix: v2.Transform = classification.get_mixup_cutmix(0.5, 5, True)
        self.assertIsInstance(mixup_cutmix, v2.Transform)
        self.assertEqual(len(mixup_cutmix.transforms), 3)  # identity, mixup, cutmix
        self.assertIsInstance(mixup_cutmix.transforms[0], v2.Identity)

        mixup_cutmix = classification.get_mixup_cutmix(None, 5, False)
        self.assertIsInstance(mixup_cutmix, v2.Transform)
        self.assertEqual(len(mixup_cutmix.transforms), 1)  # Only identity
        self.assertIsInstance(mixup_cutmix.transforms[0], v2.Identity)

        # Mixup module
        mixup = classification.RandomMixup(5, 0.2, 1.0)
        samples, targets = mixup(torch.rand((2, 3, 96, 96)), torch.tensor([0, 1], dtype=torch.int64))
        self.assertSequenceEqual(targets.size(), (2, 5))
        self.assertSequenceEqual(samples.size(), (2, 3, 96, 96))
        repr(mixup)

        # Presets
        classification.training_preset((256, 256), "birder", 0, classification.get_rgb_stats("none"))
        classification.training_preset((256, 256), "birder", 8, classification.get_rgb_stats("birder"))
        classification.training_preset((256, 256), "3aug", 3, classification.get_rgb_stats("none"))
        classification.inference_preset((256, 256), classification.get_rgb_stats("none"), 0.9)
        classification.inference_preset((256, 256), classification.get_rgb_stats("none"), 0.9, True)

    def test_detection(self) -> None:
        # Presets
        detection.training_preset((256, 256), "birder", 0, classification.get_rgb_stats("none"), False, False)
        detection.training_preset((256, 256), "birder", 5, classification.get_rgb_stats("birder"), False, True)
        detection.training_preset((256, 256), "ssd", 0, classification.get_rgb_stats("none"), True, False)
        detection.training_preset((256, 256), "multiscale", 0, classification.get_rgb_stats("none"), False, False)
        detection.InferenceTransform((256, 256), classification.get_rgb_stats("birder"), False)

        # Multiscale
        sizes = detection.build_multiscale_sizes()
        self.assertEqual(sizes, (480, 512, 544, 576, 608, 640, 672, 704, 736, 768, 800))
        if len(sizes) > 1:
            self.assertEqual(sizes[1] - sizes[0], detection.MULTISCALE_STEP)

        self.assertEqual(detection.build_multiscale_sizes(481, max_size=513), (512,))
        self.assertEqual(detection.build_multiscale_sizes(500, max_size=620), (512, 544, 576, 608))


class TestMosaic(unittest.TestCase):
    def _create_dummy_data(self) -> tuple[list[Image.Image], list[dict[str, Any]]]:
        images = []
        targets = []
        for i in range(4):
            # Create 100x100 images with different colors
            img = Image.new("RGB", (100, 100), color=(i * 50, 100, 100))
            images.append(img)

            # Create a box covering the center 50x50 area
            boxes = torch.tensor([[25.0, 25.0, 75.0, 75.0]], dtype=torch.float32)
            labels = torch.tensor([i + 1], dtype=torch.int64)

            # Wrap in tv_tensors as the pipeline expects
            boxes = tv_tensors.BoundingBoxes(boxes, format=tv_tensors.BoundingBoxFormat.XYXY, canvas_size=(100, 100))
            targets.append({"boxes": boxes, "labels": labels})

        return (images, targets)

    def test_mosaic_random_center(self) -> None:
        images, targets = self._create_dummy_data()
        output_size = (300, 300)

        out_img, out_target = mosaic.mosaic_random_center(images, targets, output_size, fill_value=(114, 114, 114))

        self.assertEqual(out_img.size, output_size)
        self.assertIsInstance(out_target["boxes"], tv_tensors.BoundingBoxes)
        self.assertEqual(out_target["boxes"].canvas_size, (output_size[1], output_size[0]))  # H, W

        # Verify boxes are within bounds
        if len(out_target["boxes"]) > 0:
            self.assertTrue((out_target["boxes"][:, 0] >= 0).all())
            self.assertTrue((out_target["boxes"][:, 1] >= 0).all())
            self.assertTrue((out_target["boxes"][:, 2] <= output_size[0]).all())
            self.assertTrue((out_target["boxes"][:, 3] <= output_size[1]).all())

        # Empty targets
        empty_targets = [{"boxes": torch.zeros((0, 4)), "labels": torch.zeros((0,))} for _ in range(4)]
        _, out_target_empty = mosaic.mosaic_random_center(images, empty_targets, output_size, fill_value=0)
        self.assertEqual(len(out_target_empty["boxes"]), 0)
        self.assertEqual(len(out_target_empty["labels"]), 0)

    def test_mosaic_fixed_grid(self) -> None:
        images, targets = self._create_dummy_data()
        output_size = (300, 300)

        # Crop to square
        out_img, out_target = mosaic.mosaic_fixed_grid(
            images, targets, output_size, fill_value=114, crop_to_square=True
        )

        self.assertEqual(out_img.size, output_size)
        self.assertIsInstance(out_target["boxes"], tv_tensors.BoundingBoxes)

        # Aspect ratio limit
        out_img_ar, _ = mosaic.mosaic_fixed_grid(
            images, targets, output_size, fill_value=114, crop_to_square=False, max_aspect_ratio=1.5
        )
        self.assertEqual(out_img_ar.size, output_size)

        # Missing keys handling
        empty_targets: list[dict[str, Any]] = [{} for _ in range(4)]
        _, out_target_empty = mosaic.mosaic_fixed_grid(images, empty_targets, output_size, fill_value=0)
        self.assertEqual(out_target_empty["boxes"].shape, (0, 4))



"""
NeWT benchmark, adapted from
https://github.com/samuelstevens/biobench/blob/main/src/biobench/newt/__init__.py

Paper "Benchmarking Representation Learning for Natural World Image Collections",
https://arxiv.org/abs/2103.16483
"""

# Reference license: MIT

import argparse
import logging
import time
from pathlib import Path
from typing import Any

import numpy as np
import numpy.typing as npt
import polars as pl
from rich.console import Console
from rich.table import Table

from birder.common import cli
from birder.common import lib
from birder.conf import settings
from birder.datahub.evaluation import NeWT
from birder.eval._embeddings import l2_normalize_
from birder.eval._embeddings import load_embeddings
from birder.eval.methods.simpleshot import normalize_features
from birder.eval.methods.svm import evaluate_svm

logger = logging.getLogger(__name__)

NeWTTaskData = tuple[str, npt.NDArray[np.float32], npt.NDArray[np.int_], npt.NDArray[np.float32], npt.NDArray[np.int_]]
NeWTPredictionsByCluster = dict[str, list[npt.NDArray[np.int_]]]


def _print_summary_table(results: list[dict[str, Any]]) -> None:
    console = Console()

    cluster_names = sorted({cluster for result in results for cluster in result["per_cluster_accuracy"].keys()})

    table = Table(show_header=True, header_style="bold dark_magenta")
    table.add_column("NeWT (SVM)", style="dim")
    table.add_column("Accuracy", justify="right")
    table.add_column("Std", justify="right")
    table.add_column("Runs", justify="right")
    for cluster in cluster_names:
        table.add_column(cluster.replace("_", " ").title(), justify="right")

    for result in results:
        row = [
            Path(result["embeddings_file"]).name,
            f"{result['accuracy']:.4f}",
            f"{result['accuracy_std']:.4f}",
            f"{result['num_runs']}",
        ]
        for cluster in cluster_names:
            acc = result["per_cluster_accuracy"].get(cluster)
            row.append(f"{acc:.4f}" if acc is not None else "-")

        table.add_row(*row)

    console.print(table)


def _write_results_csv(results: list[dict[str, Any]], output_path: Path) -> None:
    cluster_names = sorted({cluster for result in results for cluster in result["per_cluster_accuracy"].keys()})
    rows: list[dict[str, Any]] = []
    for result in results:
        row: dict[str, Any] = {
            "embeddings_file": result["embeddings_file"],
            "method": result["method"],
            "accuracy": result["accuracy"],
            "accuracy_std": result["accuracy_std"],
            "num_runs": result["num_runs"],
        }
        for cluster in cluster_names:
            row[f"cluster_{cluster}"] = result["per_cluster_accuracy"].get(cluster)

        rows.append(row)

    pl.DataFrame(rows).write_csv(output_path)
    logger.info(f"Results saved to {output_path}")


def _prepare_newt_task_data(embeddings_path: str, labels_df: pl.DataFrame) -> list[NeWTTaskData]:
    logger.info(f"Loading embeddings from {embeddings_path}")
    emb_df = load_embeddings(embeddings_path)

    joined = labels_df.join(emb_df, on="id", how="inner").sort("index")
    if joined.height < labels_df.height:
        logger.warning(f"Join dropped {labels_df.height - joined.height} samples (missing embeddings)")

    all_features = joined.get_column("embedding").to_numpy().astype(np.float32, copy=False)
    logger.info(f"Loaded {all_features.shape[0]} samples with {all_features.shape[1]} dimensions")

    # Global L2 normalization
    l2_normalize_(all_features)
    all_labels = joined.get_column("label").to_numpy().astype(np.int_)
    all_splits = joined.get_column("split").to_numpy()
    all_tasks = joined.get_column("task").to_numpy()
    all_clusters = joined.get_column("task_cluster").to_numpy()

    task_names = np.unique(all_tasks)
    logger.info(f"Found {len(task_names)} tasks")

    task_data: list[NeWTTaskData] = []
    for task_name in task_names:
        task_mask = all_tasks == task_name
        task_features = all_features[task_mask]
        task_labels = all_labels[task_mask]
        is_train = all_splits[task_mask] == "train"
        cluster = str(all_clusters[task_mask][0])

        x_train = task_features[is_train]
        y_train = task_labels[is_train]
        x_test = task_features[~is_train]
        y_test = task_labels[~is_train]

        if x_train.size == 0 or x_test.size == 0:
            logger.warning(f"Skipping task {task_name}: empty train or test split")
            continue

        # Per-task centering and L2 normalization
        x_train, x_test = normalize_features(x_train, x_test)
        task_data.append((cluster, x_train, y_train, x_test, y_test))

    return task_data


def _predict_newt(
    task_data: list[NeWTTaskData], n_iter: int, n_jobs: int, seed: int
) -> tuple[list[npt.NDArray[np.int_]], list[npt.NDArray[np.int_]], NeWTPredictionsByCluster, NeWTPredictionsByCluster]:
    y_preds_all: list[npt.NDArray[np.int_]] = []
    y_trues_all: list[npt.NDArray[np.int_]] = []
    cluster_preds: NeWTPredictionsByCluster = {}
    cluster_trues: NeWTPredictionsByCluster = {}

    for cluster, x_train, y_train, x_test, y_test in task_data:
        y_pred, y_true = evaluate_svm(x_train, y_train, x_test, y_test, n_iter=n_iter, n_jobs=n_jobs, seed=seed)
        y_preds_all.append(y_pred)
        y_trues_all.append(y_true)

        if cluster not in cluster_preds:
            cluster_preds[cluster] = []
            cluster_trues[cluster] = []

        cluster_preds[cluster].append(y_pred)
        cluster_trues[cluster].append(y_true)

    return (y_preds_all, y_trues_all, cluster_preds, cluster_trues)


# pylint: disable=too-many-locals
def evaluate_newt_single(
    embeddings_path: str, labels_df: pl.DataFrame, runs: int, n_iter: int, n_jobs: int, seed: int
) -> dict[str, Any]:
    task_data = _prepare_newt_task_data(embeddings_path, labels_df)

    scores: list[float] = []
    cluster_scores: dict[str, list[float]] = {}
    for run in range(runs):
        run_seed = seed + run

        y_preds_all, y_trues_all, cluster_preds, cluster_trues = _predict_newt(
            task_data, n_iter=n_iter, n_jobs=n_jobs, seed=run_seed
        )

        # Micro-averaged accuracy
        y_preds = np.concatenate(y_preds_all)
        y_trues = np.concatenate(y_trues_all)
        acc = float(np.mean(y_preds == y_trues))
        scores.append(acc)
        logger.info(f"Run {run + 1}/{runs} - Accuracy: {acc:.4f}")

        # Compute per-cluster accuracy for this run
        for cluster, preds_list in cluster_preds.items():
            preds = np.concatenate(preds_list)
            trues = np.concatenate(cluster_trues[cluster])
            cluster_acc = float(np.mean(preds == trues))
            if cluster not in cluster_scores:
                cluster_scores[cluster] = []

            cluster_scores[cluster].append(cluster_acc)

    # Average per-cluster accuracy across runs
    per_cluster_accuracy: dict[str, float] = {}
    for cluster, accs in cluster_scores.items():
        per_cluster_accuracy[cluster.lower()] = float(np.mean(accs))

    scores_arr = np.array(scores)
    mean_acc = float(scores_arr.mean())
    std_acc = float(scores_arr.std(ddof=1)) if len(scores) > 1 else 0.0

    logger.info(f"Mean accuracy over {runs} runs: {mean_acc:.4f} +/- {std_acc:.4f} (std)")
    for cluster, acc in sorted(per_cluster_accuracy.items()):
        logger.info(f"  {cluster}: {acc:.4f}")

    return {
        "method": "svm",
        "accuracy": mean_acc,
        "accuracy_std": std_acc,
        "num_runs": runs,
        "per_cluster_accuracy": per_cluster_accuracy,
        "embeddings_file": str(embeddings_path),
    }


def evaluate_newt(args: argparse.Namespace) -> None:
    tic = time.time()

    logger.info(f"Loading NeWT dataset from {args.dataset_path}")
    dataset = NeWT(args.dataset_path)
    labels_path = dataset.labels_path
    logger.info(f"Loading labels from {labels_path}")
    labels_df = pl.read_csv(labels_path).with_row_index(name="index").with_columns(pl.col("id").cast(pl.Utf8))

    results: list[dict[str, Any]] = []
    total = len(args.embeddings)
    for idx, embeddings_path in enumerate(args.embeddings, start=1):
        logger.info(f"Processing embeddings {idx}/{total}: {embeddings_path}")
        result = evaluate_newt_single(
            embeddings_path, labels_df, runs=args.runs, n_iter=args.n_iter, n_jobs=args.n_jobs, seed=args.seed
        )
        results.append(result)

    _print_summary_table(results)

    if args.dry_run is False:
        output_dir = settings.RESULTS_DIR.joinpath(args.dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        output_path = output_dir.joinpath("newt.csv")
        _write_results_csv(results, output_path)

    toc = time.time()
    logger.info(f"NeWT benchmark completed in {lib.format_duration(toc - tic)}")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "newt",
        allow_abbrev=False,
        help="run NeWT benchmark - 164 binary classification tasks evaluated with SVM",
        description="run NeWT benchmark - 164 binary classification tasks evaluated with SVM",
        epilog=(
            "Usage examples:\n"
            "python -m birder.eval newt --embeddings "
            "results/hieradet_d_small_dino-v2_0_224px_crop1.0_36032_output.parquet "
            "--dataset-path ~/Datasets/NeWT --dry-run\n"
            "python -m birder.eval newt --embeddings results/newt/*.parquet "
            "--dataset-path ~/Datasets/NeWT\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "--embeddings", type=str, nargs="+", metavar="FILE", help="paths to embeddings/logits parquet files"
    )
    subparser.add_argument("--dataset-path", type=str, metavar="PATH", help="path to NeWT dataset root")
    subparser.add_argument("--runs", type=int, default=3, help="number of evaluation runs")
    subparser.add_argument("--n-iter", type=int, default=100, help="SVM hyperparameter search iterations")
    subparser.add_argument("--n-jobs", type=int, default=8, help="parallel jobs for RandomizedSearchCV")
    subparser.add_argument("--seed", type=int, default=0, help="base random seed")
    subparser.add_argument(
        "--dir", type=str, default="newt", help="place all outputs in a sub-directory (relative to results)"
    )
    subparser.add_argument("--dry-run", default=False, action="store_true", help="skip saving results to file")
    subparser.set_defaults(func=main)


def validate_args(args: argparse.Namespace) -> None:
    if args.embeddings is None:
        raise cli.ValidationError("--embeddings is required")
    if args.dataset_path is None:
        raise cli.ValidationError("--dataset-path is required")


def main(args: argparse.Namespace) -> None:
    validate_args(args)
    evaluate_newt(args)



from collections.abc import Callable
from pathlib import Path
from typing import Optional

import numpy as np
import torch
from PIL import Image
from sklearn.decomposition import PCA

from birder.introspection.base import InterpretabilityResult
from birder.introspection.base import preprocess_image
from birder.net.base import DetectorBackbone


class FeaturePCA:
    """
    Visualizes feature maps using Principal Component Analysis

    This method extracts feature maps from a specified stage of a DetectorBackbone model,
    applies PCA to reduce the channel dimension to 3 components and visualizes them as an RGB image where:
    - R channel = 1st principal component (most important)
    - G channel = 2nd principal component
    - B channel = 3rd principal component
    """

    def __init__(
        self,
        net: DetectorBackbone,
        device: torch.device,
        transform: Callable[..., torch.Tensor],
        normalize: bool = False,
        channels_last: bool = False,
        stage: Optional[str] = None,
    ) -> None:
        self.net = net.eval()
        self.device = device
        self.transform = transform
        self.normalize = normalize
        self.channels_last = channels_last
        self.stage = stage

    def __call__(self, image: str | Path | Image.Image) -> InterpretabilityResult:
        input_tensor, rgb_img = preprocess_image(image, self.transform, self.device)

        with torch.inference_mode():
            features_dict = self.net.detection_features(input_tensor)

        if self.stage is not None:
            features = features_dict[self.stage]
        else:
            features = list(features_dict.values())[-1]  # Use the last stage by default

        features_np = features.cpu().numpy()

        # Handle channels_last format (B, H, W, C) vs channels_first (B, C, H, W)
        if self.channels_last is True:
            B, H, W, C = features_np.shape
            # Already in (B, H, W, C), just reshape to (B*H*W, C)
            features_reshaped = features_np.reshape(-1, C)
        else:
            B, C, H, W = features_np.shape
            # Reshape to (spatial_points, channels) for PCA
            features_reshaped = features_np.reshape(B, C, -1)
            features_reshaped = features_reshaped.transpose(0, 2, 1)  # (B, H*W, C)
            features_reshaped = features_reshaped.reshape(-1, C)  # (B*H*W, C)

        x = features_reshaped
        if self.normalize is True:
            x = x / (np.linalg.norm(x, axis=1, keepdims=True) + 1e-6)

        pca = PCA(n_components=3)
        pca_features = pca.fit_transform(x)
        pca_features = pca_features.reshape(B, H, W, 3)

        # Extract all 3 components (B=1)
        pca_rgb = pca_features[0]  # (H, W, 3)

        # Normalize each channel independently to [0, 1]
        for i in range(3):
            channel = pca_rgb[:, :, i]
            channel = channel - channel.min()
            channel = channel / (channel.max() + 1e-7)
            pca_rgb[:, :, i] = channel

        target_size = (input_tensor.size(-1), input_tensor.size(-2))  # PIL expects (width, height)
        pca_rgb_resized = (
            np.array(
                Image.fromarray((pca_rgb * 255).astype(np.uint8)).resize(target_size, Image.Resampling.BILINEAR)
            ).astype(np.float32)
            / 255.0
        )

        visualization = (pca_rgb_resized * 255).astype(np.uint8)

        return InterpretabilityResult(
            original_image=rgb_img,
            visualization=visualization,
            raw_output=pca_rgb.astype(np.float32),
            logits=None,
            predicted_class=None,
        )



import random
from typing import Any
from typing import Optional

import torch
from PIL import Image
from torchvision import tv_tensors


# pylint: disable=too-many-locals
def mosaic_random_center(
    images: list[Image.Image],
    targets: list[dict[str, Any]],
    output_size: tuple[int, int],
    fill_value: int | tuple[int, int, int],
    min_box_area: float = 0.0,
) -> tuple[Image.Image, dict[str, Any]]:
    """
    Create a mosaic augmentation by combining 4 images into a single image.

    This augmentation places 4 images on a canvas, meeting at a randomly selected
    center point. Each image is scaled to fit, cropped as needed and their bounding
    boxes are transformed accordingly.

    Parameters
    ----------
    images
        List of exactly 4 PIL images. Order: top-left, top-right, bottom-left, bottom-right.
    targets
        List of exactly 4 target dicts, each containing:
        - "boxes": Tensor of shape (N, 4) in XYXY format.
        - "labels": Tensor of shape (N,) with class indices.
    output_size
        Target output size as (width, height).
    fill_value
        Background fill color. Can be a single int (grayscale) or RGB tuple.
    min_box_area
        Minimum box area in pixels to keep after clipping. Boxes smaller than this are removed.

    Returns
    -------
    tuple[Image.Image, dict[str, Any]]
        The mosaic image (RGB) and merged target dict with "boxes" and "labels".
    """

    assert len(images) == 4 and len(targets) == 4

    (output_w, output_h) = output_size

    fill_color: tuple[int, int, int]
    if isinstance(fill_value, int):
        fill_color = (fill_value, fill_value, fill_value)
    else:
        fill_color = fill_value

    mosaic_img = Image.new("RGB", (output_w, output_h), color=fill_color)

    xc = int(random.uniform(output_w * 0.25, output_w * 0.75))
    yc = int(random.uniform(output_h * 0.25, output_h * 0.75))

    quadrants = [
        (0, 0, xc, yc),
        (xc, 0, output_w, yc),
        (0, yc, xc, output_h),
        (xc, yc, output_w, output_h),
    ]

    all_boxes: list[torch.Tensor] = []
    all_labels: list[torch.Tensor] = []

    for i, (img, target) in enumerate(zip(images, targets)):
        if img.mode != "RGB":
            img = img.convert("RGB")

        if "boxes" in target and len(target["boxes"]) > 0:
            boxes = target["boxes"]
            if isinstance(boxes, tv_tensors.BoundingBoxes):
                boxes = boxes.data.clone().float()
            else:
                boxes = boxes.clone().float()
            labels = target["labels"].clone()
        else:
            boxes = torch.zeros((0, 4), dtype=torch.float32)
            labels = torch.zeros((0,), dtype=torch.int64)

        (w, h) = img.size
        (q_x1, q_y1, q_x2, q_y2) = quadrants[i]

        scale = min(output_w / w, output_h / h)
        new_w = int(w * scale)
        new_h = int(h * scale)
        img = img.resize((new_w, new_h), resample=Image.Resampling.BILINEAR)

        # Canvas coordinates (where to paste)
        x1a = max(q_x1, q_x2 - new_w) if i in (0, 2) else q_x1
        y1a = max(q_y1, q_y2 - new_h) if i in (0, 1) else q_y1
        x2a = min(q_x2, x1a + new_w)
        y2a = min(q_y2, y1a + new_h)

        # Source image coordinates (what to crop)
        if i == 0:
            x1b = new_w - (x2a - x1a)
            y1b = new_h - (y2a - y1a)
        elif i == 1:
            x1b = 0
            y1b = new_h - (y2a - y1a)
        elif i == 2:
            x1b = new_w - (x2a - x1a)
            y1b = 0
        else:
            x1b = 0
            y1b = 0

        x2b = x1b + (x2a - x1a)
        y2b = y1b + (y2a - y1a)
        mosaic_img.paste(img.crop((x1b, y1b, x2b, y2b)), (x1a, y1a))

        if len(boxes) > 0:
            boxes[:, [0, 2]] *= new_w / w
            boxes[:, [1, 3]] *= new_h / h
            offset_x = x1a - x1b
            offset_y = y1a - y1b
            boxes[:, 0] += offset_x
            boxes[:, 2] += offset_x
            boxes[:, 1] += offset_y
            boxes[:, 3] += offset_y
            all_boxes.append(boxes)
            all_labels.append(labels)

    if len(all_boxes) > 0:
        boxes = torch.concat(all_boxes, dim=0)
        labels = torch.concat(all_labels, dim=0)

        boxes[:, [0, 2]] = boxes[:, [0, 2]].clamp(0, output_w)
        boxes[:, [1, 3]] = boxes[:, [1, 3]].clamp(0, output_h)

        ws = boxes[:, 2] - boxes[:, 0]
        hs = boxes[:, 3] - boxes[:, 1]
        valid = (ws > 0) & (hs > 0) & ((ws * hs) > min_box_area)
        boxes = boxes[valid]
        labels = labels[valid]

    else:
        boxes = torch.zeros((0, 4), dtype=torch.float32)
        labels = torch.zeros((0,), dtype=torch.int64)

    boxes = tv_tensors.BoundingBoxes(boxes, format=tv_tensors.BoundingBoxFormat.XYXY, canvas_size=(output_h, output_w))

    return (mosaic_img, {"boxes": boxes, "labels": labels})


# pylint: disable=too-many-locals,too-many-statements,too-many-branches
def mosaic_fixed_grid(
    images: list[Image.Image],
    targets: list[dict[str, Any]],
    output_size: tuple[int, int],
    fill_value: int | tuple[int, int, int],
    min_visibility: float = 0.05,
    crop_to_square: bool = False,
    max_aspect_ratio: Optional[float] = 2.0,
) -> tuple[Image.Image, dict[str, Any]]:
    """
    Create a grid-based mosaic augmentation by combining 4 images into a 2x2 grid.

    This augmentation arranges 4 images in a fixed 2x2 grid layout. Images are first
    resized/padded to squares, then assembled into a larger grid which is randomly
    cropped to the output size. Based on the approach from YOLOv4.

    Reference: https://github.com/prymeka/mosaic-augmentation-pytorch/blob/main/mosaic.py (MIT License)

    Parameters
    ----------
    images
        List of exactly 4 PIL images. Order: top-left, top-right, bottom-left, bottom-right.
    targets
        List of exactly 4 target dicts, each containing:
        - "boxes": Tensor of shape (N, 4) in XYXY format.
        - "labels": Tensor of shape (N,) with class indices.
    output_size
        Target output size as (width, height).
    fill_value
        Background fill color. Can be a single int (grayscale) or RGB tuple.
    min_visibility
        Minimum visibility ratio (area_after / area_before) to keep a box after cropping.
    crop_to_square
        If True, crop all images to square before assembly. Otherwise, pad to square.
    max_aspect_ratio
        Maximum aspect ratio before cropping to reduce it. Ignored if crop_to_square is True.

    Returns
    -------
    tuple[Image.Image, dict[str, Any]]
        The mosaic image (RGB) and merged target dict with "boxes" and "labels".
    """

    assert len(images) == 4 and len(targets) == 4

    (output_w, output_h) = output_size

    fill_color: tuple[int, int, int]
    if isinstance(fill_value, int):
        fill_color = (fill_value, fill_value, fill_value)
    else:
        fill_color = fill_value

    # Grid size (larger than output to allow cropping)
    min_possible_image_area = 0.25
    grid_size = int(max(output_w, output_h) * (2 - min_possible_image_area**0.5))
    image_size = grid_size // 2

    grid_img = Image.new("RGB", (grid_size, grid_size), color=fill_color)
    positions = [
        (0, 0),
        (image_size, 0),
        (0, image_size),
        (image_size, image_size),
    ]

    all_boxes: list[torch.Tensor] = []
    all_labels: list[torch.Tensor] = []

    for i, (img, target) in enumerate(zip(images, targets)):
        if img.mode != "RGB":
            img = img.convert("RGB")

        if "boxes" in target and len(target["boxes"]) > 0:
            boxes = target["boxes"]
            if isinstance(boxes, tv_tensors.BoundingBoxes):
                boxes = boxes.data.clone().float()
            else:
                boxes = boxes.clone().float()
            labels = target["labels"].clone()
        else:
            boxes = torch.zeros((0, 4), dtype=torch.float32)
            labels = torch.zeros((0,), dtype=torch.int64)

        (w, h) = img.size

        if crop_to_square:
            min_dim = min(w, h)
            left = (w - min_dim) // 2
            top = (h - min_dim) // 2
            img = img.crop((left, top, left + min_dim, top + min_dim))

            if len(boxes) > 0:
                boxes[:, 0] -= left
                boxes[:, 2] -= left
                boxes[:, 1] -= top
                boxes[:, 3] -= top
                boxes[:, [0, 2]] = boxes[:, [0, 2]].clamp(0, min_dim)
                boxes[:, [1, 3]] = boxes[:, [1, 3]].clamp(0, min_dim)

            (w, h) = img.size

        elif max_aspect_ratio is not None:
            aspect = max(w, h) / max(min(w, h), 1)
            if aspect > max_aspect_ratio:
                if w > h:
                    new_w = int(h * max_aspect_ratio)
                    if new_w < w:
                        left = (w - new_w) // 2
                        img = img.crop((left, 0, left + new_w, h))
                        if len(boxes) > 0:
                            boxes[:, 0] -= left
                            boxes[:, 2] -= left
                            boxes[:, [0, 2]] = boxes[:, [0, 2]].clamp(0, new_w)

                        (w, h) = img.size
                else:
                    new_h = int(w * max_aspect_ratio)
                    if new_h < h:
                        top = (h - new_h) // 2
                        img = img.crop((0, top, w, top + new_h))
                        if len(boxes) > 0:
                            boxes[:, 1] -= top
                            boxes[:, 3] -= top
                            boxes[:, [1, 3]] = boxes[:, [1, 3]].clamp(0, new_h)

                        (w, h) = img.size

        # Pad to square (content towards center, padding towards edges)
        max_dim = max(w, h)
        pad_w = max_dim - w
        pad_h = max_dim - h

        if i == 0:
            paste_x, paste_y = pad_w, pad_h
        elif i == 1:
            paste_x, paste_y = 0, pad_h
        elif i == 2:
            paste_x, paste_y = pad_w, 0
        else:
            paste_x, paste_y = 0, 0

        if w != h:
            padded_img = Image.new("RGB", (max_dim, max_dim), color=fill_color)
            padded_img.paste(img, (paste_x, paste_y))
            img = padded_img

            if len(boxes) > 0:
                boxes[:, 0] += paste_x
                boxes[:, 2] += paste_x
                boxes[:, 1] += paste_y
                boxes[:, 3] += paste_y

        scale = image_size / max_dim
        img = img.resize((image_size, image_size), resample=Image.Resampling.BILINEAR)

        if len(boxes) > 0:
            boxes *= scale

        (px, py) = positions[i]
        grid_img.paste(img, (px, py))

        if len(boxes) > 0:
            boxes[:, 0] += px
            boxes[:, 2] += px
            boxes[:, 1] += py
            boxes[:, 3] += py
            all_boxes.append(boxes)
            all_labels.append(labels)

    max_x = grid_size - output_w
    max_y = grid_size - output_h
    crop_x = random.randint(0, max(0, max_x))
    crop_y = random.randint(0, max(0, max_y))

    mosaic_img = grid_img.crop((crop_x, crop_y, crop_x + output_w, crop_y + output_h))

    if len(all_boxes) > 0:
        boxes = torch.concat(all_boxes, dim=0)
        labels = torch.concat(all_labels, dim=0)

        ws = boxes[:, 2] - boxes[:, 0]
        hs = boxes[:, 3] - boxes[:, 1]
        orig_areas = ws * hs

        boxes[:, 0] -= crop_x
        boxes[:, 2] -= crop_x
        boxes[:, 1] -= crop_y
        boxes[:, 3] -= crop_y
        boxes[:, [0, 2]] = boxes[:, [0, 2]].clamp(0, output_w)
        boxes[:, [1, 3]] = boxes[:, [1, 3]].clamp(0, output_h)

        ws = boxes[:, 2] - boxes[:, 0]
        hs = boxes[:, 3] - boxes[:, 1]
        valid = (ws > 0) & (hs > 0) & ((ws * hs) / (orig_areas + 1e-6) >= min_visibility)
        boxes = boxes[valid]
        labels = labels[valid]

    else:
        boxes = torch.zeros((0, 4), dtype=torch.float32)
        labels = torch.zeros((0,), dtype=torch.int64)

    boxes = tv_tensors.BoundingBoxes(boxes, format=tv_tensors.BoundingBoxFormat.XYXY, canvas_size=(output_h, output_w))

    return (mosaic_img, {"boxes": boxes, "labels": labels})



from pathlib import Path

from birder.datahub._lib import download_url
from birder.datahub._lib import extract_archive


class AwA2:
    """
    Name: AwA2 (Animals with Attributes 2)
    Link: https://cvml.ista.ac.at/AwA2/
    Size: 50 animal classes, 37,322 images, 85 binary attributes per class
    """

    attribute_names = [
        "black",
        "white",
        "blue",
        "brown",
        "gray",
        "orange",
        "red",
        "yellow",
        "patches",
        "spots",
        "stripes",
        "furry",
        "hairless",
        "toughskin",
        "big",
        "small",
        "bulbous",
        "lean",
        "flippers",
        "hands",
        "hooves",
        "pads",
        "paws",
        "longleg",
        "longneck",
        "tail",
        "chewteeth",
        "meatteeth",
        "buckteeth",
        "strainteeth",
        "horns",
        "claws",
        "tusks",
        "smelly",
        "flys",
        "hops",
        "swims",
        "tunnels",
        "walks",
        "fast",
        "slow",
        "strong",
        "weak",
        "muscle",
        "bipedal",
        "quadrapedal",
        "active",
        "inactive",
        "nocturnal",
        "hibernate",
        "agility",
        "fish",
        "meat",
        "plankton",
        "vegetation",
        "insects",
        "forager",
        "grazer",
        "hunter",
        "scavenger",
        "skimmer",
        "stalker",
        "newworld",
        "oldworld",
        "arctic",
        "coastal",
        "desert",
        "bush",
        "plains",
        "forest",
        "fields",
        "jungle",
        "mountains",
        "ocean",
        "ground",
        "water",
        "tree",
        "cave",
        "fierce",
        "timid",
        "smart",
        "group",
        "solitary",
        "nestspot",
        "domestic",
    ]

    def __init__(self, root: str | Path, download: bool = False, progress_bar: bool = True) -> None:
        if isinstance(root, str):
            root = Path(root)

        self._root = root

        if download is True:
            archive_path = root.parent.joinpath("AwA2-data.zip")
            downloaded = download_url(
                "https://cvml.ista.ac.at/AwA2/AwA2-data.zip",
                archive_path,
                sha256="cc5a849879165acaa2b52f1de3f146ffcd1c475f6ef85bab0152c763e573744f",
                progress_bar=progress_bar,
            )
            if downloaded is True or self._root.exists() is False:
                extract_archive(archive_path, root.parent)

        else:
            if self._root.exists() is False or self._root.is_dir() is False:
                raise RuntimeError("Dataset not found, try download=True to download it")

            if self.images_dir.exists() is False:
                raise RuntimeError("Dataset seems corrupted: JPEGImages directory not found")

            if self.predicate_matrix_binary_path.exists() is False:
                raise RuntimeError("Dataset seems corrupted: predicate-matrix-binary.txt not found")

    @property
    def images_dir(self) -> Path:
        return self._root.joinpath("JPEGImages")

    @property
    def classes_path(self) -> Path:
        return self._root.joinpath("classes.txt")

    @property
    def predicates_path(self) -> Path:
        return self._root.joinpath("predicates.txt")

    @property
    def predicate_matrix_binary_path(self) -> Path:
        return self._root.joinpath("predicate-matrix-binary.txt")

    @property
    def predicate_matrix_continuous_path(self) -> Path:
        return self._root.joinpath("predicate-matrix-continuous.txt")

    @property
    def trainclasses_path(self) -> Path:
        return self._root.joinpath("trainclasses.txt")

    @property
    def testclasses_path(self) -> Path:
        return self._root.joinpath("testclasses.txt")


class FishNet:
    """
    Name: FishNet
    Link: https://fishnet-2023.github.io/
    Size: 94,532 images, 17,357 aquatic species, 9 binary traits

    Traits:
        - FeedingPath (benthic=0, pelagic=1)
        - Tropical, Temperate, Subtropical, Boreal, Polar (habitat, 0/1)
        - freshwater, saltwater, brackish (water type, 0/1)

    Note: This dataset requires manual download from Google Drive.
    """

    trait_columns = [
        "FeedingPath",
        "Tropical",
        "Temperate",
        "Subtropical",
        "Boreal",
        "Polar",
        "freshwater",
        "saltwater",
        "brackish",
    ]

    def __init__(self, root: str | Path) -> None:
        if isinstance(root, str):
            root = Path(root)

        self._root = root

        # Verify dataset exists
        if self._root.exists() is False or self._root.is_dir() is False:
            raise RuntimeError(f"Dataset not found at {self._root}. Download it from https://fishnet-2023.github.io/")

        if self.images_dir.exists() is False:
            raise RuntimeError("Dataset seems corrupted: images directory not found")

        if self.train_csv.exists() is False:
            raise RuntimeError("Dataset seems corrupted: train.csv not found")

        if self.test_csv.exists() is False:
            raise RuntimeError("Dataset seems corrupted: test.csv not found")

    @property
    def images_dir(self) -> Path:
        return self._root.joinpath("images")

    @property
    def train_csv(self) -> Path:
        return self._root.joinpath("train.csv")

    @property
    def test_csv(self) -> Path:
        return self._root.joinpath("test.csv")


class FungiCLEF2023:
    """
    Name: FungiCLEF2023
    Link: https://www.imageclef.org/FungiCLEF2023
    Size: 1,604 species, ~417K images (train + val + test)
    """

    def __init__(self, root: str | Path, download: bool = False, progress_bar: bool = True) -> None:
        if isinstance(root, str):
            root = Path(root)

        self._root = root

        if download is True:
            self._root.mkdir(parents=True, exist_ok=True)

            train_images_archive = self._root.joinpath("DF20-300px.tar.gz")
            downloaded_train_images = download_url(
                "http://ptak.felk.cvut.cz/plants/DanishFungiDataset/DF20-300px.tar.gz",
                train_images_archive,
                sha256="b7b572179c3e99dfdfaed4b75872cb6cc59ad8d7dccab331906687ca6bce3b5a",
                progress_bar=progress_bar,
            )
            if downloaded_train_images is True or self.train_images_dir.exists() is False:
                extract_archive(train_images_archive, self._root)

            val_test_images_archive = self._root.joinpath("DF21_300px.tar.gz")
            downloaded_val_test_images = download_url(
                "http://ptak.felk.cvut.cz/plants/DanishFungiDataset/DF21_300px.tar.gz",
                val_test_images_archive,
                sha256="c0194d3314370a22fb01fb0800330c2e18c90d83f97def55dea84cb5abc2fc3e",
                progress_bar=progress_bar,
            )
            if downloaded_val_test_images is True or self.val_test_images_dir.exists() is False:
                extract_archive(val_test_images_archive, self._root)

            download_url(
                "http://ptak.felk.cvut.cz/plants/DanishFungiDataset/FungiCLEF2023_train_metadata_PRODUCTION.csv",
                self.train_metadata_path,
                sha256="dc17fc1ab48f0876947402965ee9c25e437c1622f134edab5c7da6c9b853d907",
                progress_bar=progress_bar,
            )
            download_url(
                "http://ptak.felk.cvut.cz/plants/DanishFungiDataset/FungiCLEF2023_val_metadata_PRODUCTION.csv",
                self.val_metadata_path,
                sha256="9573102de721bc93f36e5e03e878cd50cc7f6031a7a3bc82ed0642ec4c691c2a",
                progress_bar=progress_bar,
            )
            download_url(
                "http://ptak.felk.cvut.cz/plants/DanishFungiDataset/FungiCLEF2023_public_test_metadata_PRODUCTION.csv",
                self.test_metadata_path,
                sha256="56ae171d5abf2a99a3ccf8cd96cb685d0f96a7bda055a37afd2fda3e943d991c",
                progress_bar=progress_bar,
            )

        else:
            if self._root.exists() is False or self._root.is_dir() is False:
                raise RuntimeError("Dataset not found, try download=True to download it")

            if self.train_images_dir.exists() is False:
                raise RuntimeError("Dataset seems corrupted: DF20_300 directory not found")

            if self.val_test_images_dir.exists() is False:
                raise RuntimeError("Dataset seems corrupted: DF21_300 directory not found")

            if self.train_metadata_path.exists() is False:
                raise RuntimeError("Dataset seems corrupted: train metadata CSV not found")

            if self.val_metadata_path.exists() is False:
                raise RuntimeError("Dataset seems corrupted: validation metadata CSV not found")

            if self.test_metadata_path.exists() is False:
                raise RuntimeError("Dataset seems corrupted: test metadata CSV not found")

    @property
    def train_images_dir(self) -> Path:
        return self._root.joinpath("DF20_300")

    @property
    def val_test_images_dir(self) -> Path:
        return self._root.joinpath("DF21_300")

    @property
    def val_images_dir(self) -> Path:
        return self.val_test_images_dir

    @property
    def test_images_dir(self) -> Path:
        return self.val_test_images_dir

    @property
    def train_metadata_path(self) -> Path:
        return self._root.joinpath("FungiCLEF2023_train_metadata_PRODUCTION.csv")

    @property
    def val_metadata_path(self) -> Path:
        return self._root.joinpath("FungiCLEF2023_val_metadata_PRODUCTION.csv")

    @property
    def test_metadata_path(self) -> Path:
        return self._root.joinpath("FungiCLEF2023_public_test_metadata_PRODUCTION.csv")


class NABirds:
    """
    Name: NABirds
    Link: https://dl.allaboutbirds.org/nabirds
    Size: 555 visual categories, ~48K images

    Note: This dataset requires manual download. Visit the link above.
    """

    def __init__(self, root: str | Path) -> None:
        if isinstance(root, str):
            root = Path(root)

        self._root = root

        # Verify dataset exists
        if self._root.exists() is False or self._root.is_dir() is False:
            raise RuntimeError(
                f"Dataset not found at {self._root}. Download it from https://dl.allaboutbirds.org/nabirds"
            )

        if self.images_dir.exists() is False:
            raise RuntimeError("Dataset seems corrupted: images directory not found")

    @property
    def images_dir(self) -> Path:
        return self._root.joinpath("images")

    @property
    def images_path(self) -> Path:
        return self._root.joinpath("images.txt")

    @property
    def classes_path(self) -> Path:
        return self._root.joinpath("classes.txt")

    @property
    def labels_path(self) -> Path:
        return self._root.joinpath("image_class_labels.txt")

    @property
    def train_test_split_path(self) -> Path:
        return self._root.joinpath("train_test_split.txt")

    @property
    def hierarchy_path(self) -> Path:
        return self._root.joinpath("hierarchy.txt")


class NeWT:
    """
    Name: NeWT (Natural World Tasks)
    Link: https://github.com/visipedia/newt
    Size: 164 binary classification tasks, ~36K images
    """

    def __init__(self, root: str | Path, download: bool = False, progress_bar: bool = True) -> None:
        if isinstance(root, str):
            root = Path(root)

        self._root = root

        if download is True:
            self._root.mkdir(parents=True, exist_ok=True)

            # Download images
            images_src = root.parent.joinpath("newt2021_images.tar.gz")
            downloaded_images = download_url(
                "https://ml-inat-competition-datasets.s3.amazonaws.com/newt/newt2021_images.tar.gz",
                images_src,
                sha256="8d40958a867c1296f92b5e125f1f1d8ddaa59f249315740fc366fc606995c055",
                progress_bar=progress_bar,
            )
            if downloaded_images is True or self.images_dir.exists() is False:
                extract_archive(images_src, self._root)

            # Download labels
            labels_src = root.parent.joinpath("newt2021_labels.csv.tar.gz")
            downloaded_labels = download_url(
                "https://ml-inat-competition-datasets.s3.amazonaws.com/newt/newt2021_labels.csv.tar.gz",
                labels_src,
                sha256="e09807842485ef49ccf51d74ac9f6072c599fc16cf5ee755fdf4064f2e4c3828",
                progress_bar=progress_bar,
            )
            if downloaded_labels is True or self.labels_path.exists() is False:
                extract_archive(labels_src, self._root)

        else:
            # Some sanity checks
            if self._root.exists() is False or self._root.is_dir() is False:
                raise RuntimeError("Dataset not found, try download=True to download it")

            if self.images_dir.exists() is False:
                raise RuntimeError("Dataset seems corrupted: images directory not found")

            if self.labels_path.exists() is False:
                raise RuntimeError("Dataset seems corrupted: labels CSV not found")

    @property
    def images_dir(self) -> Path:
        return self._root.joinpath("newt2021_images")

    @property
    def labels_path(self) -> Path:
        return self._root.joinpath("newt2021_labels.csv")


class Plankton:
    """
    Name: SYKE-plankton_IFCB_2022
    Link: https://b2share.eudat.eu/records/xvnrp-7ga56
    Size: 50 phytoplankton classes, ~214K images (train + val)
    """

    def __init__(self, root: str | Path, download: bool = False, progress_bar: bool = True) -> None:
        if isinstance(root, str):
            root = Path(root)

        self._root = root

        if download is True:
            self._root.mkdir(parents=True, exist_ok=True)

            train_archive = root.parent.joinpath("phytoplankton_labeled.zip")
            downloaded_train = download_url(
                "https://b2share.eudat.eu/records/xvnrp-7ga56/files/phytoplankton_labeled.zip",
                train_archive,
                sha256="0c47acd8dfad46829fe42758a6c24adcdb5e6f2456be4ced975cbb9de9644704",
                progress_bar=progress_bar,
            )
            if downloaded_train is True or self.train_dir.exists() is False:
                extract_archive(train_archive, self._root)

            val_archive = root.parent.joinpath("phytoplankton_Uto_2021_labeled.zip")
            downloaded_val = download_url(
                "https://b2share.eudat.eu/records/w7y96-6jd66/files/phytoplankton_Ut%C3%B6_2021_labeled.zip",
                val_archive,
                sha256="b017809515c3d58171ecbfd196d6725239e9380c2a22ae880ac56e878bbfcfa4",
                progress_bar=progress_bar,
            )
            if downloaded_val is True or self.val_dir.exists() is False:
                extract_archive(val_archive, self._root)
                # Rename extracted directory to avoid non-ASCII character
                extracted_dir = self._root.joinpath("phytoplankton_Utö_2021_labeled")
                if extracted_dir.exists():
                    extracted_dir.rename(self.val_dir)

        else:
            if self._root.exists() is False or self._root.is_dir() is False:
                raise RuntimeError("Dataset not found, try download=True to download it")

            if self.train_dir.exists() is False:
                raise RuntimeError("Dataset seems corrupted: train directory not found")

            if self.val_dir.exists() is False:
                raise RuntimeError("Dataset seems corrupted: val directory not found")

    @property
    def train_dir(self) -> Path:
        return self._root.joinpath("labeled_20201020")

    @property
    def val_dir(self) -> Path:
        return self._root.joinpath("phytoplankton_Uto_2021_labeled")


class PlantDoc:
    """
    Name: PlantDoc
    Link: https://github.com/pratikkayal/PlantDoc-Dataset
    Paper: https://arxiv.org/abs/1911.10317
    Size: 27 classes (13 plant species, 17 disease categories), 2,598 images
    """

    _archive_dir_name = "PlantDoc-Dataset-5467f6012d78d1c446145d5f582da6096f852ae8"

    def __init__(self, root: str | Path, download: bool = False, progress_bar: bool = True) -> None:
        if isinstance(root, str):
            root = Path(root)

        self._root = root

        if download is True:
            archive_path = root.parent.joinpath("plantdoc.zip")
            downloaded = download_url(
                "https://github.com/pratikkayal/PlantDoc-Dataset/archive/"
                "5467f6012d78d1c446145d5f582da6096f852ae8.zip",
                archive_path,
                sha256="94e2b99a500a63efbd48923ed48588fbb01f9b1db66a2d3b5c24eed6466da20f",
                progress_bar=progress_bar,
            )
            if downloaded is True or self._root.exists() is False:
                extract_archive(archive_path, root.parent)
                # Rename extracted directory from commit hash name to friendly name
                extracted_dir = root.parent.joinpath(self._archive_dir_name)
                extracted_dir.rename(self._root)

        else:
            if self._root.exists() is False or self._root.is_dir() is False:
                raise RuntimeError("Dataset not found, try download=True to download it")

            if self.train_dir.exists() is False:
                raise RuntimeError("Dataset seems corrupted: train directory not found")

            if self.test_dir.exists() is False:
                raise RuntimeError("Dataset seems corrupted: test directory not found")

    @property
    def train_dir(self) -> Path:
        return self._root.joinpath("train")

    @property
    def test_dir(self) -> Path:
        return self._root.joinpath("test")


class PlantNet:
    """
    Name: PlantNet-300K
    Link: https://plantnet.org/en/2021/03/30/a-plntnet-dataset-for-machine-learning-researchers/
    Size: 1081 species, ~300K images
    """

    def __init__(self, root: str | Path, download: bool = False, progress_bar: bool = True) -> None:
        if isinstance(root, str):
            root = Path(root)

        self._root = root

        if download is True:
            archive_path = root.parent.joinpath("plantnet_300K.zip")
            downloaded = download_url(
                "https://zenodo.org/records/5645731/files/plantnet_300K.zip?download=1",
                archive_path,
                sha256="3a079076c8ad4476beac54d89ea344958256a999428937eba47ec352dadce00d",
                progress_bar=progress_bar,
            )
            if downloaded is True or self._root.exists() is False:
                extract_archive(archive_path, root.parent)

        else:
            if self._root.exists() is False or self._root.is_dir() is False:
                raise RuntimeError("Dataset not found, try download=True to download it")

            if self.images_dir.exists() is False:
                raise RuntimeError("Dataset seems corrupted: images directory not found")

            if self.species_id_to_name_path.exists() is False:
                raise RuntimeError("Dataset seems corrupted: species_id_2_name.json not found")

    @property
    def images_dir(self) -> Path:
        return self._root.joinpath("images")

    @property
    def train_dir(self) -> Path:
        return self.images_dir.joinpath("train")

    @property
    def val_dir(self) -> Path:
        return self.images_dir.joinpath("val")

    @property
    def test_dir(self) -> Path:
        return self.images_dir.joinpath("test")

    @property
    def species_id_to_name_path(self) -> Path:
        return self._root.joinpath("plantnet300K_species_id_2_name.json")

    @property
    def metadata_path(self) -> Path:
        return self._root.joinpath("plantnet300K_metadata.json")



"""
Attention Rollout for Vision Transformers, adapted from
https://github.com/jacobgil/vit-explain/blob/main/vit_rollout.py

Paper "Quantifying Attention Flow in Transformers", https://arxiv.org/abs/2005.00928
"""

# Reference license: MIT

from collections.abc import Callable
from pathlib import Path
from typing import Literal
from typing import Optional

import numpy as np
import torch
from PIL import Image
from torch import nn

from birder.introspection.base import InterpretabilityResult
from birder.introspection.base import predict_class
from birder.introspection.base import preprocess_image
from birder.introspection.base import show_mask_on_image
from birder.net.vit import Encoder


# pylint: disable=too-many-locals
def compute_rollout(
    attentions: list[torch.Tensor],
    discard_ratio: float,
    head_fusion: Literal["mean", "max", "min"],
    num_special_tokens: int,
    patch_grid_shape: tuple[int, int],
) -> torch.Tensor:
    # Assume batch size = 1
    num_tokens = attentions[0].size(-1)
    device = attentions[0].device

    # Start with identity (residual)
    result = torch.eye(num_tokens, device=device)

    with torch.no_grad():
        for attention in attentions:
            # Fuse heads: [B, H, T, T] -> [B, T, T]
            if head_fusion == "mean":
                attention_heads_fused = attention.mean(dim=1)
            elif head_fusion == "max":
                attention_heads_fused = attention.max(dim=1)[0]
            elif head_fusion == "min":
                attention_heads_fused = attention.min(dim=1)[0]
            else:
                raise ValueError(f"Unsupported head_fusion: {head_fusion}")

            # attention_heads_fused: [1, T, T] (batch = 1)
            if discard_ratio > 0:
                # Work on the single batch element
                attn = attention_heads_fused[0]  # [T, T]

                # Define which positions are "non-special"
                idx = torch.arange(num_tokens, device=attn.device)
                is_special = idx < num_special_tokens
                non_special = ~is_special

                # We are only allowed to prune NON-special <-> NON-special entries
                allow = non_special[:, None] & non_special[None, :]  # [T, T]

                allowed_values = attn[allow]
                num_allowed = allowed_values.numel()
                if num_allowed > 0:
                    num_to_discard = int(num_allowed * discard_ratio)
                    if num_to_discard > 0:
                        # Drop the smallest allowed values
                        _, low_idx = torch.topk(allowed_values, num_to_discard, largest=False)
                        allowed_values[low_idx] = 0
                        attn[allow] = allowed_values
                        attention_heads_fused[0] = attn

            # Add residual connection and normalize
            eye = torch.eye(num_tokens, device=attention_heads_fused.device)
            a = (attention_heads_fused + eye) / 2.0  # [1, T, T]
            a = a / a.sum(dim=-1, keepdim=True)

            # Accumulate attention across layers
            result = torch.matmul(a, result)  # [1, T, T]

    rollout = result[0]  # [T, T]

    # Build final token → patch map
    if 0 < num_special_tokens < num_tokens:
        # Sources: all special tokens (0 .. num_special_tokens-1)
        # Targets: all non-special tokens (num_special_tokens .. end)
        source_to_patches = rollout[:num_special_tokens, num_special_tokens:]
        mask = source_to_patches.mean(dim=0)
    else:
        # No special tokens (or all are special): fall back to averaging over all sources
        mask = rollout.mean(dim=0)  # [T]

    # Normalize and reshape to 2D map using actual patch grid dimensions
    mask = mask / (mask.max() + 1e-8)
    grid_h, grid_w = patch_grid_shape
    mask = mask.reshape(grid_h, grid_w)

    return mask


class AttentionGatherer:
    def __init__(self, net: nn.Module, attention_layer_name: str) -> None:
        assert hasattr(net, "encoder") is True and isinstance(net.encoder, Encoder)

        net.encoder.set_need_attn()
        self.net = net
        self.attentions: list[torch.Tensor] = []
        self.handles: list[torch.utils.hooks.RemovableHandle] = []

        # Register hooks on attention layers
        for name, module in self.net.named_modules():
            if name.endswith(attention_layer_name) is True:
                handle = module.register_forward_hook(self._capture_attention)
                self.handles.append(handle)

    def _capture_attention(
        self, _module: nn.Module, _inputs: tuple[torch.Tensor, ...], outputs: tuple[torch.Tensor, ...]
    ) -> None:
        self.attentions.append(outputs[1].cpu())

    def __call__(self, x: torch.Tensor) -> tuple[list[torch.Tensor], torch.Tensor]:
        self.attentions = []
        with torch.inference_mode():
            logits = self.net(x)

        return (self.attentions, logits)

    def release(self) -> None:
        for handle in self.handles:
            handle.remove()


class AttentionRollout:
    def __init__(
        self,
        net: nn.Module,
        device: torch.device,
        transform: Callable[..., torch.Tensor],
        attention_layer_name: str = "attn",
        discard_ratio: float = 0.9,
        head_fusion: Literal["mean", "max", "min"] = "max",
    ) -> None:
        if not 0 <= discard_ratio <= 1:
            raise ValueError(f"discard_ratio must be in [0, 1], got {discard_ratio}")

        self.net = net.eval()
        self.device = device
        self.transform = transform
        self.discard_ratio = discard_ratio
        self.head_fusion = head_fusion
        self.attention_gatherer = AttentionGatherer(net, attention_layer_name)

    def __call__(self, image: str | Path | Image.Image, target_class: Optional[int] = None) -> InterpretabilityResult:
        input_tensor, rgb_img = preprocess_image(image, self.transform, self.device)

        attentions, logits = self.attention_gatherer(input_tensor)

        _, _, H, W = input_tensor.shape
        patch_grid_shape = (H // self.net.stem_stride, W // self.net.stem_stride)

        attention_map = compute_rollout(
            attentions, self.discard_ratio, self.head_fusion, self.net.num_special_tokens, patch_grid_shape
        )
        attention_img = Image.fromarray(attention_map.numpy())
        attention_img = attention_img.resize((rgb_img.shape[1], rgb_img.shape[0]))
        attention_arr = np.array(attention_img)

        visualization = show_mask_on_image(rgb_img, attention_arr, image_weight=0.4)

        return InterpretabilityResult(
            original_image=rgb_img,
            visualization=visualization,
            raw_output=attention_arr,
            logits=logits.detach(),
            predicted_class=predict_class(logits),
        )

    def __del__(self) -> None:
        if hasattr(self, "attention_gatherer") is True:
            self.attention_gatherer.release()



"""
Flowers102 benchmark using SimpleShot for flower species classification

Paper "Automated Flower Classification over a Large Number of Classes"
https://www.robots.ox.ac.uk/~vgg/data/flowers/102/
"""

import argparse
import logging
import time
from pathlib import Path
from typing import Any

import numpy as np
import numpy.typing as npt
import polars as pl
from rich.console import Console
from rich.table import Table
from torchvision.datasets import ImageFolder

from birder.common import cli
from birder.common import lib
from birder.conf import settings
from birder.eval._embeddings import load_embeddings
from birder.eval.methods.simpleshot import evaluate_simpleshot

logger = logging.getLogger(__name__)


def _print_summary_table(results: list[dict[str, Any]]) -> None:
    console = Console()

    table = Table(show_header=True, header_style="bold dark_magenta")
    table.add_column("Flowers102 (SimpleShot)", style="dim")
    table.add_column("Val Acc", justify="right")
    table.add_column("Test Acc", justify="right")

    for result in results:
        table.add_row(
            Path(result["embeddings_file"]).name,
            f"{result['val_accuracy']:.4f}",
            f"{result['test_accuracy']:.4f}",
        )

    console.print(table)


def _write_results_csv(results: list[dict[str, Any]], output_path: Path) -> None:
    rows: list[dict[str, Any]] = []
    for result in results:
        rows.append(
            {
                "embeddings_file": result["embeddings_file"],
                "method": result["method"],
                "val_accuracy": result["val_accuracy"],
                "test_accuracy": result["test_accuracy"],
            }
        )

    pl.DataFrame(rows).write_csv(output_path)
    logger.info(f"Results saved to {output_path}")


def _load_flowers102_metadata(dataset_path: Path) -> pl.DataFrame:
    rows: list[dict[str, Any]] = []
    for split in ["training", "validation", "testing"]:
        split_dir = dataset_path.joinpath(split)
        if not split_dir.exists():
            continue

        dataset = ImageFolder(str(split_dir))
        for path, label in dataset.samples:
            rows.append({"id": Path(path).stem, "label": label, "split": split})

    return pl.DataFrame(rows)


def _load_embeddings_with_split(embeddings_path: str, metadata_df: pl.DataFrame) -> tuple[
    npt.NDArray[np.float32],
    npt.NDArray[np.int_],
    npt.NDArray[np.float32],
    npt.NDArray[np.int_],
    npt.NDArray[np.float32],
    npt.NDArray[np.int_],
]:
    logger.info(f"Loading embeddings from {embeddings_path}")
    emb_df = load_embeddings(embeddings_path)

    train_meta = metadata_df.filter(pl.col("split") == "training").select(["id", "label"])
    val_meta = metadata_df.filter(pl.col("split") == "validation").select(["id", "label"])
    test_meta = metadata_df.filter(pl.col("split") == "testing").select(["id", "label"])

    train_join = train_meta.join(emb_df, on="id", how="inner")
    val_join = val_meta.join(emb_df, on="id", how="inner")
    test_join = test_meta.join(emb_df, on="id", how="inner")

    dropped_train = train_meta.height - train_join.height
    dropped_val = val_meta.height - val_join.height
    dropped_test = test_meta.height - test_join.height
    dropped_total = dropped_train + dropped_val + dropped_test
    if dropped_total > 0:
        logger.warning(
            f"Join dropped {dropped_total} samples (missing embeddings): "
            f"train={dropped_train}, val={dropped_val}, test={dropped_test}"
        )

    x_train = train_join.get_column("embedding").to_numpy().astype(np.float32, copy=False)
    y_train = train_join.get_column("label").to_numpy().astype(np.int_)
    x_val = val_join.get_column("embedding").to_numpy().astype(np.float32, copy=False)
    y_val = val_join.get_column("label").to_numpy().astype(np.int_)
    x_test = test_join.get_column("embedding").to_numpy().astype(np.float32, copy=False)
    y_test = test_join.get_column("label").to_numpy().astype(np.int_)

    num_classes = metadata_df.get_column("label").max() + 1  # type: ignore[operator]
    total_samples = x_train.shape[0] + x_val.shape[0] + x_test.shape[0]
    embedding_dim = x_train.shape[1]
    logger.info(f"Loaded {total_samples} samples with {embedding_dim} dimensions, {num_classes} classes")

    logger.info(f"Train: {len(y_train)}, Val: {len(y_val)}, Test: {len(y_test)} samples")

    return (x_train, y_train, x_val, y_val, x_test, y_test)


def evaluate_flowers102_single(
    x_train: npt.NDArray[np.float32],
    y_train: npt.NDArray[np.int_],
    x_val: npt.NDArray[np.float32],
    y_val: npt.NDArray[np.int_],
    x_test: npt.NDArray[np.float32],
    y_test: npt.NDArray[np.int_],
    embeddings_path: str,
) -> dict[str, Any]:
    # Evaluate on validation set
    y_pred_val, y_true_val = evaluate_simpleshot(x_train, y_train, x_val, y_val)
    val_acc = float(np.mean(y_pred_val == y_true_val))
    logger.info(f"Validation accuracy: {val_acc:.4f}")

    # Evaluate on test set
    y_pred_test, y_true_test = evaluate_simpleshot(x_train, y_train, x_test, y_test)
    test_acc = float(np.mean(y_pred_test == y_true_test))
    logger.info(f"Test accuracy: {test_acc:.4f}")

    return {
        "method": "simpleshot",
        "val_accuracy": val_acc,
        "test_accuracy": test_acc,
        "embeddings_file": str(embeddings_path),
    }


def evaluate_flowers102(args: argparse.Namespace) -> None:
    tic = time.time()

    logger.info(f"Loading Flowers102 dataset from {args.dataset_path}")
    dataset_path = Path(args.dataset_path)
    metadata_df = _load_flowers102_metadata(dataset_path)
    logger.info(f"Loaded metadata for {metadata_df.height} images")

    results: list[dict[str, Any]] = []
    total = len(args.embeddings)
    for idx, embeddings_path in enumerate(args.embeddings, start=1):
        logger.info(f"Processing embeddings {idx}/{total}: {embeddings_path}")
        x_train, y_train, x_val, y_val, x_test, y_test = _load_embeddings_with_split(embeddings_path, metadata_df)

        result = evaluate_flowers102_single(x_train, y_train, x_val, y_val, x_test, y_test, embeddings_path)
        results.append(result)

    _print_summary_table(results)

    if args.dry_run is False:
        output_dir = settings.RESULTS_DIR.joinpath(args.dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        output_path = output_dir.joinpath("flowers102.csv")
        _write_results_csv(results, output_path)

    toc = time.time()
    logger.info(f"Flowers102 benchmark completed in {lib.format_duration(toc - tic)}")


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "flowers102",
        allow_abbrev=False,
        help="run Flowers102 benchmark - 102 class classification using SimpleShot",
        description="run Flowers102 benchmark - 102 class classification using SimpleShot",
        epilog=(
            "Usage examples:\n"
            "python -m birder.eval flowers102 --embeddings "
            "results/flowers102_embeddings.parquet --dataset-path ~/Datasets/Flowers102 --dry-run\n"
            "python -m birder.eval flowers102 --embeddings results/flowers102/*.parquet "
            "--dataset-path ~/Datasets/Flowers102\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument(
        "--embeddings", type=str, nargs="+", metavar="FILE", help="paths to embeddings parquet files"
    )
    subparser.add_argument("--dataset-path", type=str, metavar="PATH", help="path to Flowers102 dataset root")
    subparser.add_argument(
        "--dir", type=str, default="flowers102", help="place all outputs in a sub-directory (relative to results)"
    )
    subparser.add_argument("--dry-run", default=False, action="store_true", help="skip saving results to file")
    subparser.set_defaults(func=main)


def validate_args(args: argparse.Namespace) -> None:
    if args.embeddings is None:
        raise cli.ValidationError("--embeddings is required")
    if args.dataset_path is None:
        raise cli.ValidationError("--dataset-path is required")


def main(args: argparse.Namespace) -> None:
    validate_args(args)
    evaluate_flowers102(args)
